import asyncio
import binascii
import logging
from abc import abstractmethod

from paradox.config import config as cfg
from paradox.connections.handler import ConnectionHandler, IPConnectionHandler
from paradox.connections.ip.parsers import (IPMessageCommand, IPMessageRequest,
                                            IPMessageResponse, IPMessageType)

logger = logging.getLogger("PAI").getChild(__name__)


def checksum(data, min_message_length):
    """Calculates the 8bit checksum of Paradox messages"""
    c = 0

    if data is None or len(data) < min_message_length:
        return False

    for i in data[:-1]:
        c += i

    r = (c % 256) == data[-1]
    return r


class ConnectionProtocol(asyncio.Protocol):
    def __init__(self, handler: ConnectionHandler):
        self.transport = None
        self.use_variable_message_length = True
        self.buffer = b""

        self.handler = handler

        self._closed = asyncio.get_event_loop().create_future()
        self.buffer = b""

    def connection_made(self, transport):
        self.transport = transport

        self.handler.on_connection()

    def is_active(self) -> bool:
        return bool(self.transport) and not self._closed.done()

    def check_active(self):
        if not self.is_active():
            raise ConnectionError("Transport does not exist or is already closed")

    async def close(self):
        if self.transport:
            try:
                self.transport.close()
            except:
                logger.exception("Connection transport close raised Exception")
            self.transport = None

        await asyncio.wait_for(self._closed, timeout=1)

    @abstractmethod
    def send_message(self, message):
        raise NotImplementedError("This function needs to be overridden in a subclass")

    def connection_lost(self, exc):
        logger.error(f"Connection was closed: {exc}")
        self.buffer = b""
        self.transport = None

        if not self._closed.done():
            if exc is None:
                self._closed.set_result(None)
            else:
                self._closed.set_exception(exc)

        super().connection_lost(exc)

        # asyncio.get_event_loop().call_soon(self.on_con_lost)
        self.handler.on_connection_loss()

        self.handler = None

    def variable_message_length(self, mode):
        self.use_variable_message_length = mode

    def __del__(self):
        # Prevent reports about unhandled exceptions.
        # Better than self._closed._log_traceback = False hack
        closed = self._closed
        if closed.done() and not closed.cancelled():
            closed.exception()


class SerialConnectionProtocol(ConnectionProtocol):
    def send_message(self, message):
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("PAI -> SER {}".format(binascii.hexlify(message)))

        self.check_active()

        self.transport.write(message)

    def data_received(self, recv_data):
        self.buffer += recv_data

        min_length = 4 if self.use_variable_message_length else 37

        while len(self.buffer) >= min_length:
            if self.use_variable_message_length:
                if self.buffer[0] >> 4 == 0:
                    potential_packet_length = 37
                elif self.buffer[0] >> 4 in [1, 3, 4, 5, 6, 7, 8, 9]:
                    potential_packet_length = (
                        self.buffer[1] if 0 < self.buffer[1] <= 71 else 37
                    )
                elif self.buffer[0] >> 4 in [0x0A, 0x0B, 0x0D]:
                    potential_packet_length = self.buffer[1]
                elif self.buffer[0] >> 4 == 0x0C:
                    potential_packet_length = self.buffer[1] * 256 + self.buffer[2]
                elif self.buffer[0] >> 4 == 0x0E:
                    if self.buffer[1] < 37 or self.buffer[1] == 0xFF:
                        # MG/SP in 21st century and EVO Live Events. Probable values=0x13, 0x13, 0x00, 0xFF
                        potential_packet_length = 37
                    else:
                        potential_packet_length = self.buffer[1]
                else:
                    potential_packet_length = 37

            else:
                potential_packet_length = 37

            if len(self.buffer) < potential_packet_length:
                break

            frame = self.buffer[:potential_packet_length]

            if checksum(frame, min_length):
                self.buffer = self.buffer[len(frame) :]  # Remove message
                if cfg.LOGGING_DUMP_PACKETS:
                    logger.debug("SER -> PAI {}".format(binascii.hexlify(frame)))

                self.handler.on_message(frame)
            else:
                self.buffer = self.buffer[1:]


class IPConnectionProtocol(ConnectionProtocol):
    def __init__(self, handler: IPConnectionHandler, key):
        super(IPConnectionProtocol, self).__init__(handler)

        self.handler = handler
        self.key = key

    def send_raw(self, raw):
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("PAI -> IP (raw) {}".format(binascii.hexlify(raw)))

        self.check_active()

        self.transport.write(raw)

    def send_message(self, message):
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("PAI -> IP (payload) {}".format(binascii.hexlify(message)))

        self.check_active()

        msg = IPMessageRequest.build(
            dict(
                header=dict(
                    length=len(message),
                    message_type=IPMessageType.serial_passthrough_request,
                    flags=dict(installer_mode=True),
                    command=IPMessageCommand.passthrough,
                    wt=100,
                    cryptor_code="aes_256_ecb",
                ),
                payload=message,
            ),
            password=self.key,
        )
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("PAI -> IP (raw) {}".format(binascii.hexlify(msg)))

        self.transport.write(msg)

    def _process_message(self, data):
        message = IPMessageResponse.parse(data, password=self.key)

        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug(
                "IP -> PAI (payload) {}".format(binascii.hexlify(message.payload))
            )

        if message.header.message_type == IPMessageType.serial_passthrough_response:
            self.handler.on_message(message.payload)
        elif message.header.message_type == IPMessageType.ip_response:
            self.handler.on_ip_message(message)
        else:
            logger.error(f"Wrong message detected: {message}")

    def data_received(self, recv_data):
        self.buffer += recv_data

        if self.buffer[0] != 0xAA:
            if len(self.buffer) > 0:
                logger.warning(
                    "Dangling data in the receive buffer: %s"
                    % binascii.hexlify(self.buffer)
                )
            self.buffer = b""
            return

        if len(recv_data) + 16 < self.buffer[1]:
            return

        if len(self.buffer) % 16 != 0:
            return

        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("IP -> PAI (raw) {}".format(binascii.hexlify(self.buffer)))

        self._process_message(self.buffer)
        self.buffer = b""



import asyncio
import logging
import signal
import sys
from logging.handlers import RotatingFileHandler

from paradox import VERSION
from paradox.config import config as cfg
from paradox.exceptions import PAICriticalException
from paradox.interfaces.interface_manager import InterfaceManager
from paradox.lib.encodings import register_encodings
from paradox.paradox import Paradox

alarm = None
interface_manager = None

logger = logging.getLogger("PAI")


def get_format(level):
    if level <= logging.DEBUG:
        return (
            "%(asctime)s - %(levelname)-8s - %(threadName)-10s - %(name)s - %(message)s"
        )
    else:
        return "%(asctime)s - %(levelname)-8s - %(name)s - %(message)s"


def config_logger(logger):
    logger_level = cfg.LOGGING_LEVEL_CONSOLE

    if cfg.LOGGING_FILE is not None:
        logfile_handler = RotatingFileHandler(
            cfg.LOGGING_FILE,
            mode="a",
            maxBytes=cfg.LOGGING_FILE_MAX_SIZE * 1024 * 1024,
            backupCount=cfg.LOGGING_FILE_MAX_FILES,
            encoding=None,
            delay=0,
        )

        logfile_handler.setLevel(cfg.LOGGING_LEVEL_FILE)
        logfile_handler.setFormatter(logging.Formatter(get_format(logger_level)))
        logger.addHandler(logfile_handler)
        logger_level = min(logger_level, cfg.LOGGING_LEVEL_FILE)

    logconsole_handler = logging.StreamHandler()
    logconsole_handler.setLevel(cfg.LOGGING_LEVEL_CONSOLE)
    logconsole_handler.setFormatter(logging.Formatter(get_format(logger_level)))
    logger.addHandler(logconsole_handler)

    logger.setLevel(logger_level)


async def exit_handler(signame=None):
    global alarm, interface_manager

    if signame is not None:
        logger.info(f"Captured signal {signame}. Exiting")

    if alarm:
        await alarm.disconnect()
        alarm = None

    if interface_manager:
        interface_manager.stop()
        interface_manager = None

    logger.info("Good bye!")


async def run_loop():
    retry = 1
    while alarm is not None:
        logger.info("Starting...")
        retry_time_wait = 2 ^ retry
        retry_time_wait = 30 if retry_time_wait > 30 else retry_time_wait

        try:
            if await alarm.full_connect():
                retry = 1
                await alarm.loop()
            else:
                logger.error("Unable to connect to alarm")

            if alarm:
                await asyncio.sleep(retry_time_wait)
        except ConnectionError as e:  # Connection to IP Module or MQTT lost
            logger.error("Connection to panel lost: %s. Restarting" % str(e))
            await asyncio.sleep(retry_time_wait)
        except OSError:  # Connection to IP Module or MQTT lost
            logger.exception("Restarting")
            await asyncio.sleep(retry_time_wait)
        except PAICriticalException:
            logger.exception("PAI Critical exception. Stopping PAI")
            break
        except (KeyboardInterrupt, SystemExit):
            break  # break exits the retry loop
        except:
            logger.exception("Restarting")
            await asyncio.sleep(retry_time_wait)

        retry += 1

    if alarm:
        await exit_handler()


def main(args):
    global alarm, interface_manager

    if "config" in args and args.config is not None:
        import os

        config_file = os.path.abspath(args.config)
        cfg.load(config_file)
    else:
        cfg.load()

    config_logger(logger)

    logger.info(f"Starting Paradox Alarm Interface {VERSION}")
    logger.info(f"Config loaded from {cfg.CONFIG_FILE_LOCATION}")

    logger.info(f"Console Log level set to {cfg.LOGGING_LEVEL_CONSOLE}")

    # Registering additional encodings
    register_encodings()

    # Start interacting with the alarm
    alarm = Paradox()
    loop = asyncio.get_event_loop()
    for signame in ("SIGINT", "SIGTERM"):
        sig = getattr(signal, signame)
        loop.add_signal_handler(
            sig, lambda: asyncio.ensure_future(exit_handler(signame))
        )

    interface_manager = InterfaceManager(alarm, config=cfg)
    interface_manager.start()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_loop())

    sys.exit(0)



from paradox.event import EventLevel, LiveEvent


def _toggle_provider(old):
    return not old


def _minor2_provider(event, *_):
    return event.minor2


def _request_status_refresh(alarm, *_, **__):
    alarm.request_status_refresh()


def _zone_generated_alarm(event, storage, *_, **__):
    assert isinstance(event, LiveEvent)
    storage.update_container_object(
        "partition",
        event.partition,
        {
            "current_state": "triggered",
            "alarm_in_memory": True,
            "was_in_alarm": True,
            "audible_alarm": True,
            "exit_delay": False,
        },
    )


event_map = {
    0: dict(
        level=EventLevel.DEBUG,
        change=dict(open=False),
        type="zone",
        message="Zone {label} OK",
    ),
    1: dict(
        level=EventLevel.DEBUG,
        change=dict(open=True),
        type="zone",
        message="Zone {label} open",
    ),
    2: dict(
        level=EventLevel.WARN,
        tags=["trouble"],
        change=dict(zone_tamper_trouble=True),
        type="zone",
        message="Zone {label} tampered",
    ),
    3: dict(
        level=EventLevel.WARN,
        tags=["trouble"],
        type="zone",
        change=dict(fire_loop_trouble=True),
        message="Zone {label} in fire loop trouble",
    ),
    4: dict(
        level=EventLevel.DEBUG,
        message="Non-reportable event",
        sub={
            0: dict(
                level=EventLevel.CRITICAL,
                type="system",
                tags=["trouble"],
                message="Telephone line trouble",
            ),
            1: dict(type="system", message="Smoke detector reset"),
            2: dict(
                level=EventLevel.INFO,
                tags=["arm"],
                type="partition",
                change=dict(arm=True),
                message="Arm with no entry delay",
            ),
            3: dict(
                level=EventLevel.INFO,
                tags=["arm"],
                type="partition",
                change=dict(arm_stay=True, arm=True),
                message="Arm in stay mode",
            ),
            4: dict(
                level=EventLevel.INFO,
                tags=["arm"],
                type="partition",
                change=dict(arm_away=True, arm=True),
                message="Arm in Away mode",
            ),
            5: dict(
                level=EventLevel.INFO,
                tags=["arm"],
                type="partition",
                change=dict(arm_stay=True, arm=True),
                message="Full arm when in Stay mode",
            ),
            6: dict(type="system", message="Voice module access"),
            7: dict(
                type="system",
                message="Remote control access by {@user}",
                user=_minor2_provider,
            ),
            8: dict(type="system", message="PC fail to communicate"),
            9: dict(type="system", message="Midnight"),
            10: dict(type="system", message="Neware user login"),
            11: dict(type="system", message="Neware user logout"),
            12: dict(type="system", message="User initiated call-up"),
            13: dict(type="system", message="Force answer"),
            14: dict(type="system", message="Force hangup"),
            15: dict(type="system", message="Reset to default"),
            16: dict(type="system", message="Auxiliary output manually activated"),
            17: dict(type="system", message="Auxiliary output manually deactivated"),
            18: dict(type="system", message="Voice reporting failed"),
            19: dict(
                type="system",
                message="Fail to communicate restore",
                telephone=_minor2_provider,
            ),
            20: dict(
                type="system", message="Software access (VDMP3, Ip100, Neware, WinLoad)"
            ),  # minor2 has sub events
            21: dict(type="system", message="IPR512 1 registration status"),
            22: dict(type="system", message="IPR512 2 registration status"),
            23: dict(type="system", message="IPR512 3 registration status"),
            24: dict(type="system", message="IPR512 4 registration status"),
        },
    ),
    5: dict(  # minor = user id
        level=EventLevel.INFO,
        type="user",
        message="User {label} code entered on keypad",
        keypad=_minor2_provider,
    ),
    6: dict(
        level=EventLevel.INFO,
        tags=["access"],
        type="door",
        message="{@user} access on door {label}",
        user=_minor2_provider,
    ),
    7: dict(level=EventLevel.INFO, type="user", message="Bypass programming access"),
    8: dict(level=EventLevel.WARN, type="zone", message="TX delay zone {label} alarm"),
    9: dict(  # minor = user id, label in message = user name
        level=EventLevel.INFO,
        tags=["arm"],
        change=dict(arm=True),
        type="partition",
        message="Arming {label} with {@user} master code",
    ),
    10: dict(  # minor = user id, label in message = user name
        level=EventLevel.INFO,
        tags=["arm"],
        change=dict(arm=True),
        type="partition",
        message="Arming {label} with {@user} code",
    ),
    11: dict(
        level=EventLevel.INFO,
        tags=["arm"],
        change=dict(arm=True),
        type="partition",
        message="Arming {label} with keyswitch",
    ),
    12: dict(
        level=EventLevel.INFO,
        tags=["arm"],
        change=dict(arm=True),
        type="partition",
        message="Special arming {label}",
        sub={  # no label passed
            0: dict(type="partition", message="auto arming"),
            1: dict(
                type="partition",
                message="arming with Winload by {@user}",
                user=_minor2_provider,
            ),
            2: dict(type="partition", message="late to close"),
            3: dict(type="partition", message="no movement arming"),
            4: dict(type="partition", message="partial arming"),
            5: dict(type="partition", message="one-touch arming"),
            6: dict(type="partition", message="Future use 1"),
            7: dict(type="partition", message="Future use 2"),
            8: dict(type="partition", message="(InTouch) voice module arming"),
            9: dict(type="partition", message="Delinquency c;psomg"),
            10: dict(type="partition", message="Future use 3"),
        },
    ),
    13: dict(  # minor = user id, label in message = user name
        level=EventLevel.INFO,
        tags=["disarm"],
        change=dict(arm=False, current_state="disarmed", audible_alarm=False),
        type="partition",
        message="{label} disarmed with {@user} master code",
    ),
    14: dict(  # minor = user id, label in message = user name
        level=EventLevel.INFO,
        tags=["disarm"],
        change=dict(arm=False, current_state="disarmed", audible_alarm=False),
        type="partition",
        message="{label} disarmed with {@user} code",
    ),
    15: dict(
        level=EventLevel.INFO,
        tags=["disarm"],
        change=dict(arm=False, current_state="disarmed", audible_alarm=False),
        type="partition",
        message="{label} disarmed with keyswitch",
    ),
    16: dict(
        level=EventLevel.INFO,
        tags=["disarm", "after_alarm"],
        change=dict(arm=False, current_state="disarmed", audible_alarm=False),
        type="partition",
        message="{label} disarmed after alarm with master",
    ),
    17: dict(
        level=EventLevel.INFO,
        tags=["disarm", "after_alarm"],
        change=dict(arm=False, current_state="disarmed", audible_alarm=False),
        type="partition",
        message="{label} disarmed after alarm with user code",
    ),
    18: dict(
        level=EventLevel.INFO,
        tags=["disarm", "after_alarm"],
        change=dict(arm=False, current_state="disarmed", audible_alarm=False),
        type="partition",
        message="{label} disarmed after alarm with keyswitch",
    ),
    19: dict(
        level=EventLevel.INFO,
        tags=["alarm", "cancel"],
        type="user",
        message="Alarm cancelled by {label} (master)",
    ),
    20: dict(
        level=EventLevel.INFO,
        tags=["alarm", "cancel"],
        type="user",
        message="Alarm cancelled by {label} (user code)",
    ),
    21: dict(
        level=EventLevel.INFO,
        tags=["alarm", "cancel"],
        type="keyswitch",
        message="{label} Alarm cancelled with keyswitch",
    ),
    22: dict(
        level=EventLevel.INFO,
        tags=["disarm"],
        change=dict(arm=False),
        type="partition",
        message="{label} special disarming",
        sub={  # no label passed
            0: dict(type="partition", message="Auto arm cancelled"),
            1: dict(type="partition", message="One-touch stay/instant disarm"),
            2: dict(
                type="partition",
                message="Disarming with Winload by {@user}",
                user=_minor2_provider,
            ),
            3: dict(
                type="partition",
                tags=["after_alarm"],
                message="Disarming with Winload after alarm by {@user}",
                user=_minor2_provider,
            ),
            4: dict(
                type="partition",
                message="Winload cancelled alarm by {@user}",
                user=_minor2_provider,
            ),
            5: dict(type="partition", message="Future use"),
            6: dict(type="partition", message="Future use"),
            7: dict(type="partition", message="Future use"),
            8: dict(type="partition", message="(InTouch) voice module disarming"),
            9: dict(type="partition", message="Future use"),
        },
    ),
    23: dict(
        level=EventLevel.INFO,
        type="zone",
        change=dict(bypassed=_toggle_provider),
        tags=["bypass"],
        message="Zone {label} bypass toggled",
    ),
    24: dict(
        level=EventLevel.CRITICAL,
        tags=["alarm", "trigger"],
        type="zone",
        change=dict(generated_alarm=True, presently_in_alarm=True),
        message="Zone {label} in alarm",
        hook_fn=_zone_generated_alarm,
    ),
    25: dict(
        level=EventLevel.CRITICAL,
        tags=["alarm", "fire", "trigger"],
        type="zone",
        change=dict(fire_alarm=True),
        message="Zone {label} in fire alarm",
        hook_fn=_request_status_refresh,
    ),
    26: dict(
        level=EventLevel.WARN,
        tags=["alarm", "restore"],
        type="zone",
        change=dict(presently_in_alarm=False),
        message="Zone {label} alarm restore",
        hook_fn=_request_status_refresh,
    ),
    27: dict(
        level=EventLevel.WARN,
        tags=["alarm", "fire", "restore"],
        type="zone",
        change=dict(fire_alarm=False),
        message="Zone {label} fire alarm restore",
        hook_fn=_request_status_refresh,
    ),
    28: dict(
        level=EventLevel.INFO, type="user", message="{label} Early to disarm by user"
    ),
    29: dict(
        level=EventLevel.INFO, type="user", message="{label} Late to disarm by user"
    ),
    30: dict(
        level=EventLevel.CRITICAL,
        type="special",
        message="Special alarm",
        tags=["alarm", "trigger"],
        sub={
            0: dict(message="Panic emergency"),
            1: dict(message="Panic medical"),
            2: dict(message="Panic fire"),
            3: dict(message="Recent closing"),
            4: dict(message="Police code"),
            5: dict(message="Zone shutdown"),
            6: dict(message="Future use 1"),
            7: dict(message="Future use 2"),
            8: dict(message="TLM alarm"),
            9: dict(message="Central communication failure alarm"),
            10: dict(message="Module tamper alarm"),
            11: dict(message="Missing GSM module alarm"),
            12: dict(message="GSM no service alarm"),
            13: dict(message="Missing IP module alarm"),
            14: dict(message="IP no service alarm"),
            15: dict(message="Missing voice module alarm"),
        },
        hook_fn=_request_status_refresh,
    ),
    31: dict(
        level=EventLevel.CRITICAL,
        type="user",
        tags=["alarm", "trigger"],
        message="Duress alarm by user {label}",
        hook_fn=_request_status_refresh,
    ),
    32: dict(
        level=EventLevel.INFO,
        type="zone",
        change=dict(shutdown=True),
        message="Zone {label} shutdown",
    ),
    33: dict(
        level=EventLevel.INFO,
        tags=["tamper"],
        type="zone",
        change=dict(zone_tamper_trouble=True),
        message="Zone {label} tamper",
    ),
    34: dict(
        level=EventLevel.INFO,
        tags=["tamper", "restore"],
        type="zone",
        change=dict(zone_tamper_trouble=True),
        message="Zone {label} tamper restore",
    ),
    35: dict(
        level=EventLevel.INFO,
        tags=["tamper"],
        type="zone",
        message="Zone special {label} tamper",
        sub={0: dict(message="Keypad Lockout"), 1: dict(message="Voice lockout")},
    ),
    36: dict(
        level=EventLevel.WARN,
        type="system",
        message="Trouble",
        tags=["trouble"],
        sub={
            0: dict(message="TLM trouble"),
            1: dict(message="AC failure"),
            2: dict(message="Battery failure"),
            3: dict(message="Auxiliary current overload"),
            4: dict(message="Bell current overload"),
            5: dict(message="Bell disconnected"),
            6: dict(tags=["clock"], message="Clock loss"),
            7: dict(message="Fire loop trouble"),
            8: dict(message="Fail to communicate to monitoring station telephone #1"),
            9: dict(message="Fail to communicate to monitoring station telephone #2"),
            11: dict(message="Fail to communicate to voice report"),
            12: dict(message="RF jamming"),
            13: dict(message="GSM RF jamming"),
            14: dict(message="GSM no service"),
            15: dict(message="GSM supervision lost"),
            16: dict(message="Fail To Communicate IP Receiver 1 (GPRS)"),
            17: dict(message="Fail To Communicate IP Receiver 2 (GPRS)"),
            18: dict(message="IP Module No Service"),
            19: dict(message="IP Module Supervision Loss"),
            20: dict(message="Fail To Communicate IP Receiver 1 (IP)"),
            21: dict(message="Fail To Communicate IP Receiver 2 (IP)"),
        },
    ),
    37: dict(
        level=EventLevel.INFO,
        type="system",
        message="Trouble restore",
        tags=["trouble", "restore"],
        sub={
            0: dict(message="TLM trouble restore"),
            1: dict(message="AC failure restore"),
            2: dict(message="Battery failure restore"),
            3: dict(message="Auxiliary current overload restore"),
            4: dict(message="Bell current overload restore"),
            5: dict(message="Bell disconnected restore"),
            6: dict(tags=["clock"], message="Clock loss restore"),
            7: dict(message="Fire loop trouble restore"),
            8: dict(
                message="Fail to communicate to monitoring station telephone #1 restore"
            ),
            9: dict(
                message="Fail to communicate to monitoring station telephone #2 restore"
            ),
            11: dict(message="Fail to communicate to voice report restore"),
            12: dict(message="RF jamming restore"),
            13: dict(message="GSM RF jamming restore"),
            14: dict(message="GSM no service restore"),
            15: dict(message="GSM supervision lost restore"),
            16: dict(message="Fail To Communicate IP Receiver 1 (GPRS) restore"),
            17: dict(message="Fail To Communicate IP Receiver 2 (GPRS) restore"),
            18: dict(message="IP Module No Service restore"),
            19: dict(message="IP Module Supervision Loss restore"),
            20: dict(message="Fail To Communicate IP Receiver 1 (IP) restore"),
            21: dict(message="Fail To Communicate IP Receiver 2 (IP) restore"),
            99: dict(message="Any trouble event restore"),
        },
    ),
    38: dict(
        level=EventLevel.WARN,
        type="system",
        message="Module trouble",
        tags=["trouble"],
        sub={
            0: dict(message="Combus fault"),
            1: dict(message="Module tamper"),
            2: dict(message="ROM/RAM error"),
            3: dict(message="TLM trouble"),
            4: dict(message="Fail to communicate"),
            5: dict(message="Printer fault"),
            6: dict(message="AC failure"),
            7: dict(message="Battery failure"),
            8: dict(message="Auxiliary failure"),
            9: dict(message="Future use"),
        },
        module=_minor2_provider,
    ),
    39: dict(
        level=EventLevel.INFO,
        type="system",
        message="Module trouble restore",
        tags=["trouble", "restore"],
        sub={
            0: dict(message="Combus fault restored"),
            1: dict(message="Module tamper restored"),
            2: dict(message="ROM/RAM error restored"),
            3: dict(message="TLM trouble restored"),
            4: dict(message="Fail to communicate restored"),
            5: dict(message="Printer fault restored"),
            6: dict(message="AC failure restored"),
            7: dict(message="Battery failure restored"),
            8: dict(message="Auxiliary failure restored"),
            9: dict(message="Future use"),
        },
        module=_minor2_provider,
    ),
    40: dict(
        level=EventLevel.WARN,
        type="system",
        message="Fail to communicate on telephone number",
        telephone=_minor2_provider,
    ),
    41: dict(
        level=EventLevel.WARN,
        type="zone",
        changes=dict(zone_low_battery_trouble=True),
        tags=["trouble"],
        message="Low battery on zone {label}",
    ),
    42: dict(
        level=EventLevel.WARN,
        type="zone",
        changes=dict(zone_supervision_trouble=True),
        tags=["trouble"],
        message="Zone {label} supervision trouble",
    ),
    43: dict(
        level=EventLevel.INFO,
        type="zone",
        changes=dict(zone_low_battery_trouble=False),
        tags=["trouble", "restore"],
        message="Low battery on zone {label} restored",
    ),
    44: dict(
        level=EventLevel.INFO,
        type="zone",
        changes=dict(zone_supervision_trouble=False),
        tags=["trouble", "restore"],
        message="Zone {label} supervision trouble restored",
    ),
    45: dict(
        level=EventLevel.INFO,
        type="system",
        message="Special events",
        sub={
            0: "Power-up after total power down",
            1: "Software reset (watchdog)",
            2: "Test report",
            3: "Listen-in request",
            4: "WinLoad in (connected)",
            5: "WinLoad out (disconnected)",
            6: "Installer in programming",
            7: "Installer out of programming",
            8: "Future use",
        },
    ),
    46: dict(level=EventLevel.WARN, type="user", message="Early to arm by user"),
    47: dict(level=EventLevel.WARN, type="user", message="Late to arm by user"),
    48: dict(level=EventLevel.INFO, type="utility", message="Utility key"),
    49: dict(level=EventLevel.WARN, type="door", message="Request for exit"),
    50: dict(level=EventLevel.WARN, type="door", message="Access denied"),
    51: dict(level=EventLevel.WARN, type="door", message="Door left open alarm"),
    52: dict(level=EventLevel.WARN, type="door", message="Door forced alarm"),
    53: dict(level=EventLevel.WARN, type="door", message="Door left open restore"),
    54: dict(level=EventLevel.WARN, type="door", message="Door forced open restore"),
    55: dict(level=EventLevel.WARN, type="zone", message="Intellizone triggered"),
    56: dict(
        level=EventLevel.WARN, type="zone", message="Zone excluded on Force arming"
    ),
    57: dict(
        level=EventLevel.WARN, type="zone", message="Zone went back to arm status"
    ),
    58: dict(
        level=EventLevel.INFO, type="module", message="New module assigned on combus"
    ),
    59: dict(
        level=EventLevel.INFO,
        type="module",
        message="Module manually removed from combus",
    ),
    60: dict(
        level=EventLevel.INFO,
        type="system",
        message="Non-saved event",
        sub={0: "Remote control rejected", 1: "Future use"},
    ),
    61: dict(
        level=EventLevel.INFO,
        type="module",
        message="Module manually removed from combus",
    ),
    62: dict(
        level=EventLevel.INFO,
        type="user",
        tags=["access", "granted"],
        message="{@door} access granted to user {label}",
        door=_minor2_provider,
    ),
    63: dict(
        level=EventLevel.INFO,
        type="user",
        tags=["access", "denied"],
        message="{@door} access denied to user {label}",
        door=_minor2_provider,
    ),
    64: dict(
        level=EventLevel.INFO,
        type="system",
        message="Status 1",
        sub={
            0: "Armed",
            1: "Force armed",
            2: "Stay armed",
            3: "Instant armed",
            4: "Strobe alarm",
            5: "Silent alarm",
            6: "Audible alarm",
            7: "Fire alarm",
        },
    ),
    65: dict(
        level=EventLevel.INFO,
        type="system",
        message="Status 2",
        sub={
            0: "Ready",
            1: "Exit delay",
            2: "Entry delay",
            3: "System in trouble",
            4: "Alarm in memory",
            5: "Zones bypassed",
            6: "Bypass, master, installer programming",
            7: "Keypad lockout",
        },
    ),
    66: dict(
        level=EventLevel.INFO,
        type="system",
        message="Status 3",
        sub={
            0: "Intellizone delay engaged",
            1: "Fire delay engaged",
            2: "Auto arm",
            3: "Arming with voice module (set until exit delay finishes)",
            4: "Tamper",
            5: "Zone low battery",
            6: "Fire loop trouble",
            7: "Zone supervision trouble",
        },
    ),
    67: dict(
        level=EventLevel.INFO,
        type="system",
        message="Special status",
        sub={
            # 0-3: Chime in partition 1-4
            4: "Smoke detector power reset",
            5: "Ground start",
            6: "Kiss off",
            7: "Telephone ring",
            # 8-15: Bell on partition 1-8
            # 16-23: Pulsed alarm in partition
            # 24-31: Open/close Kiss off in partition
            # 32-63: Keyswitch/PGM input
            # 64-95: Status of access door
            96: "Trouble in system",
            97: "Trouble in dialer",
            98: "Trouble in module",
            99: "Trouble in combus",
            103: "Time and date trouble",
            104: "AC failure",
            105: "Battery failure",
            106: "Auxiliary current limit",
            107: "Bell current limit",
            108: "Bell absent",
            109: "ROM error",
            110: "RAM error",
            111: "Future use",
            112: "TLM 1 trouble",
            113: "Fail to communicate 1",
            114: "Fail to communicate 2",
            115: "Fail to communicate 3",
            116: "Fail to communicate 4",
            117: "Fail to communicate with PC",
            120: "Module tamper trouble",
            121: "Module ROM error",
            122: "Module TLM error",
            123: "Module Failure to communicate",
            124: "Module printer trouble",
            125: "Module AC failure",
            126: "Module battery trouble",
            127: "Module auxiliary failure",
            128: "Missing  keypad",
            129: "Missing  module",
            133: "Global combus failure",
            134: "Combus overload",
            136: "Dialer relay",
        },
    ),
}



from paradox.event import EventLevel

property_map = {
    "ac_failure_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "power"],
        message={"True": "AC Power failure", "False": "AC Power restored"},
    ),
    "ac_loss_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "power"],
        message={
            "True": "{Type} {label} AC Power lost",
            "False": "{Type} {label} AC Power is OK",
        },
    ),
    "alarm": dict(
        level=EventLevel.CRITICAL,
        tags=["alarm"],
        message={
            "True": "{Type} {label} in alarm",
            "False": "{Type} {label} not in alarm",
        },
    ),
    "alarms_in_memory": dict(
        level=EventLevel.INFO,
        tags=["alarm"],
        message={
            "True": "{Type} {label} has alarms in memory",
            "False": "{Type} {label} alarms memory empty",
        },
    ),
    "arm": dict(
        level=EventLevel.INFO,
        tags=["arm"],
        message={
            "True": "{Type} {label} is armed",
            "False": "{Type} {label} is disarmed",
        },
    ),
    "arm_force": dict(
        level=EventLevel.INFO,
        tags=["arm"],
        message={
            "True": "{Type} {label} forced arm",
            "False": "{Type} {label} arm not forced",
        },
    ),
    "arm_sleep": dict(
        level=EventLevel.INFO,
        tags=["arm"],
        message={
            "True": "{Type} {label} sleep arm is active",
            "False": "{Type} {label} sleep arm is inactive",
        },
    ),
    "arm_stay": dict(
        level=EventLevel.INFO,
        tags=["arm"],
        message={
            "True": "{Type} {label} stay arm active",
            "False": "{Type} {label} stay arm inactive",
        },
    ),
    "arm_with_remote": dict(
        level=EventLevel.INFO,
        tags=["arm"],
        message={
            "True": "{Type} {label} is armed with remote",
            "False": "{Type} {label} is not armed with remote",
        },
    ),
    "audible_alarm": dict(
        level=EventLevel.CRITICAL,
        tags=["alarm"],
        message={
            "True": "{Type} {label} alarm is audible",
            "False": "{Type} {label} alarm is silent",
        },
    ),
    "auto_arming_engaged": dict(
        level=EventLevel.CRITICAL,
        tags=["arm"],
        message={
            "True": "{Type} {label} auto arming engaged",
            "False": "{Type} {label} auto arming is not engaged",
        },
    ),
    "auxiliary_output_overload_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "output"],
        message={
            "True": "Auxiliary output overloaded",
            "False": "Auxiliary output load is adequate",
        },
    ),
    "battery": dict(
        level=EventLevel.DEBUG,
        tags=["voltage", "battery"],
        message="Battery voltage is {value}V",
    ),
    "battery_failure_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["battery", "trouble"],
        message={
            "True": "{Type} {label} battery is low",
            "False": "{Type} {label} battery is OK",
        },
    ),
    "bell_activated": dict(
        level=EventLevel.CRITICAL,
        tags=["bell", "alarm"],
        message={
            "True": "{Type} {label} bell is active",
            "False": "{Type} {label} bell is inactive",
        },
    ),
    "bell_delay_finished": dict(
        level=EventLevel.CRITICAL,
        tags=["bell", "alarm"],
        message={
            "True": "{Type} {label} bell delay is inactive",
            "False": "{Type} {label} bell delay is active",
        },
    ),
    "bell_output_disconnected_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "bell"],
        message={"True": "Bell output disconnected", "False": "Bell output connected"},
    ),
    "bell_output_overload_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "bell"],
        message={
            "True": "Bell output overloaded",
            "False": "Bell output load is adequate",
        },
    ),
    "bell_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "bell"],
        message={"True": "Bell trouble", "False": "Bell restored"},
    ),
    "bypassed": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} bypassed",
            "False": "{Type} {label} not bypassed",
        },
    ),
    "central_1_reporting_ftc_indicator_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "central"],
        message={
            "True": "Central 1 reporting trouble",
            "False": "Central 1 reporting OK",
        },
    ),
    "central_2_reporting_ftc_indicator_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "central"],
        message={
            "True": "Central 2 reporting trouble",
            "False": "Central 2 reporting OK",
        },
    ),
    "communication_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "communication"],
        message={"True": "Communication trouble", "False": "Communication restored"},
    ),
    "computer_fail_to_communicate_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "communication"],
        message={"True": "PC communication failed", "False": "PC communication OK"},
    ),
    "dc": dict(
        level=EventLevel.DEBUG,
        tags=["voltage", "power"],
        message="DC voltage is {value}V",
    ),
    "entry_delay": dict(
        level=EventLevel.INFO,
        tags=["entry", "delay"],
        message={
            "True": "{Type} {label} in entry delay",
            "False": "{Type} {label} not in entry delay",
        },
    ),
    "entry_delay_finished": dict(
        level=EventLevel.DEBUG,
        tags=["entry", "delay"],
        message={
            "True": "{Type} {label} entry delay is finished",
            "False": "{Type} {label} entry delay is not finished",
        },
    ),
    "exit_delay": dict(
        level=EventLevel.INFO,
        tags=["exit", "delay"],
        message={
            "True": "{Type} {label} in exit delay",
            "False": "{Type} {label} not in exit delay",
        },
    ),
    "exit_delay_finished": dict(
        level=EventLevel.DEBUG,
        tags=["exit", "delay"],
        message={
            "True": "{Type} {label} exit delay is finished",
            "False": "{Type} {label} exit delay is not finished",
        },
    ),
    "fire_delay": dict(
        level=EventLevel.INFO,
        tags=["alarm", "fire"],
        message={
            "True": "{Type} {label} is in fire delay",
            "False": "{Type} {label} is in fire delay",
        },
    ),
    "fire": dict(
        level=EventLevel.CRITICAL,
        tags=["alarm", "fire"],
        message={"True": "{Type} {label} fire", "False": "{Type} {label} has no fire"},
    ),
    "fire_loop_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "Fire loop", "False": "Fire loop recovered"},
    ),
    "in_remote_delay": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} in remote delay",
            "False": "{Type} {label} not in remote delay",
        },
    ),
    "intellizone_delay": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} in intellizone delay",
            "False": "{Type} {label} not in intellizone delay",
        },
    ),
    "intellizone_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} intellizone delay finished",
            "False": "{Type} {label} intellizone delay is active",
        },
    ),
    "in_tx_delay": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} in TX delay",
            "False": "{Type} {label} not in TX delay",
        },
    ),
    "keypad_battery_failure_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["battery", "trouble"],
        message={
            "True": "Keypad {l} battery is low",
            "False": "Keypad {label} battery is OK",
        },
    ),
    "low_battery_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["battery", "trouble"],
        message={
            "True": "{Type} {label} Battery is low",
            "False": "{Type} {label} Battery level is adequate",
        },
    ),
    "module_supervision_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["supervision", "trouble"],
        message={
            "True": "Module supervision trouble",
            "False": "Module supervision OK",
        },
    ),
    "module_tamper_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["tamper", "trouble"],
        message={"True": "Module tampered", "False": "Module tamper cleared"},
    ),
    "no_delay": dict(
        level=EventLevel.CRITICAL,
        message={
            "True": "{Type} {label} has no delay",
            "False": "{Type} {label} has a delay active",
        },
    ),
    "open": dict(
        level=EventLevel.DEBUG,
        message={"True": "Zone {label} open", "False": "Zone {label} closed"},
    ),
    "pager_fail_to_communicate_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["pager", "communication", "trouble"],
        message={
            "True": "Pager communication failed",
            "False": "Pager communication OK",
        },
    ),
    "paramedic_alarm": dict(
        level=EventLevel.CRITICAL,
        tags=["alarm", "medic"],
        message={
            "True": "{Type} {label} paramedic alarm active",
            "False": "{Type} {label} paramedic alarm is inactive",
        },
    ),
    "power_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "power"],
        message={"True": "Power trouble", "False": "Power restored"},
    ),
    "pulse_fire_alarm": dict(
        level=EventLevel.CRITICAL,
        tags=["fire", "alarm"],
        message={
            "True": "{Type} {label} fire alarm pulse is active",
            "False": "{Type} {label} fire alarm pulse is inactive",
        },
    ),
    "ready_status": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} ready", "False": "{Type} {label} not ready"},
    ),
    "recent_closing_delay": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} recent closing delay",
            "False": "{Type} {label} no recent closing delay",
        },
    ),
    "rf_interference_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "rf", "interference"],
        message={"True": "RF Interference trouble", "False": "RF Interference cleared"},
    ),
    "rf_low_battery_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble", "rf", "battery"],
        message={
            "True": "{Type} {label} RF battery low",
            "False": "{Type} {label} RF battery OK",
        },
    ),
    "noise_floor": dict(
        level=EventLevel.DEBUG,
        tags=["rf", "interference"],
        message="RF Noise floor is {value}",
    ),
    "rf_supervision_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["rf", "supervision", "trouble"],
        message={
            "True": "{Type} {label} has supervision trouble",
            "False": "{Type} {label} has no supervision trouble",
        },
    ),
    "shutdown": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} is shutdown",
            "False": "{Type} {label} is active",
        },
    ),
    "signal_strength": dict(
        level=EventLevel.DEBUG,
        tags=["rf"],
        message="{Type} {label} signal strength {value}",
    ),
    "silent_alarm": dict(
        level=EventLevel.CRITICAL,
        tags=["alarm"],
        message={
            "True": "{Type} {label} silent alarm is active",
            "False": "{Type} {label} silent alarm is inactive",
        },
    ),
    "squawk": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} Squawk On",
            "False": "{Type} {label} Squawk Off",
        },
    ),
    "stayd_mode_active": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} StayD mode is active",
            "False": "{Type} {label} StayD mode is inactive",
        },
    ),
    "strobe_alarm": dict(
        level=EventLevel.CRITICAL,
        tags=["alarm"],
        message={
            "True": "{Type} {label} Alarm Strobe is active",
            "False": "{Type} {label} Alarm Strobe is inactive",
        },
    ),
    "supervision_failure_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["supervision", "trouble"],
        message={
            "True": "{Type} {label} supervision lost",
            "False": "{Type} {label} supervision is OK",
        },
    ),
    "supervision_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["supervision", "trouble"],
        message={
            "True": "{Type} {label} supervision trouble",
            "False": "{Type} {label} supervision OK",
        },
    ),
    "tamper": dict(
        level=EventLevel.CRITICAL,
        tags=["tamper", "trouble"],
        message={
            "True": "{Type} {label} tampered",
            "False": "{Type} {label} tamper cleared",
        },
    ),
    "telephone_line": dict(
        level=EventLevel.CRITICAL,
        tags=["telephone", "communication", "trouble"],
        message={"True": "Telephone line trouble", "False": "Telephone line OK"},
    ),
    "timer_loss_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["timer", "trouble"],
        message={"True": "Timer lost trouble", "False": "Timer recovered"},
    ),
    "transmission_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} Transmission delay finished",
            "False": "{Type} {label} Transmission delay is active",
        },
    ),
    "trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "{Type} {label} has trouble",
            "False": "No trouble for {Type} {label}",
        },
    ),
    "vdc": dict(
        level=EventLevel.DEBUG,
        tags=["voltage", "power"],
        message="VDC voltage is {value}V",
    ),
    "voice_fail_to_communicate_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["voice", "communication", "trouble"],
        message={
            "True": "Voice module communication failed",
            "False": "Voice module communication OK",
        },
    ),
    "wait_window": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} wait window is active",
            "False": "{Type} {label} wait window is inactive",
        },
    ),
    "was_bypassed": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} was bypassed",
            "False": "{Type} {label} was not bypassed",
        },
    ),
    "was_in_alarm": dict(
        level=EventLevel.INFO,
        tags=["alarm"],
        message={
            "True": "{Type} {label} was in alarm",
            "False": "{Type} {label} was not in alarm",
        },
    ),
    "wireless_keypad_ac_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["ac", "power", "trouble"],
        message={
            "True": "Wireless Keypad AC power with trouble",
            "False": "Wireless Keypad AC power OK",
        },
    ),
    "wireless_keypad_battery_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["battery", "trouble"],
        message={
            "True": "Wireless Keypad battery with trouble",
            "False": "Wireless Keypad battery OK",
        },
    ),
    "wireless_repeater_ac_loss_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["ac", "power", "trouble"],
        message={
            "True": "Wireless Repeater with AC power failure",
            "False": "Wireless Repeater AC power is OK",
        },
    ),
    "wireless_repeater_battery_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["battery", "trouble"],
        message={
            "True": "Wireless Repeater battery with trouble",
            "False": "Wireless Repeater battery OK",
        },
    ),
    "zone_bypassed": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} is bypassed",
            "False": "{Type} {label} is not bypassed",
        },
    ),
    "zone_supervision_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["supervision", "trouble"],
        message={
            "True": "Zone with supervision trouble",
            "False": "Zone supervision is OK",
        },
    ),
    "zone_tamper_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["tamper", "trouble"],
        message={"True": "Zone tampered", "False": "No Zone is tampered"},
    ),
    "time": dict(
        level=EventLevel.DEBUG, tags=["system", "time"], message="Panel time is {value}"
    ),
    # ---------------------
    # Synthetic properties
    # ---------------------
    "current_state": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "triggered": "{Type} {label} triggered alarm",
            "pending": "{Type} {label} arming pending",
            "armed_home": "{Type} {label} armed stay",
            "armed_away": "{Type} {label} armed away",
            "disarmed": "{Type} {label} disarmed",
        },
    ),
}



import binascii
import logging

from paradox.connections.ip.parsers import (IPMessageCommand, IPMessageRequest,
                                            IPPayloadConnectResponse)
from paradox.exceptions import ConnectToIpModuleFailed, PAICriticalException

logger = logging.getLogger("PAI").getChild(__name__)


class IPModuleConnectCommand:
    def __init__(self, connection):
        self.connection = connection

    async def execute(self):
        await self._authenticate_to_ip_module()

        await self._send_keep_alive_command()

        # # F4
        # logger.debug("Sending F4")
        # msg = binascii.unhexlify('aa00000309f400000001eeeeeeee0000')
        # self.connection.send_raw(msg)
        # message_payload = await self.wait_for_message(raw=True)
        #
        # logger.debug("F4 answer: {}".format(binascii.hexlify(message_payload)))

        await self._send_upload_download_connection_command()

        # F8
        await self._send_toggle_keep_alive_command()

        logger.info("Session successfully established with IP Module")

    async def _send_toggle_keep_alive_command(self):
        logger.debug("Sending toggle keep alive request")
        payload = binascii.unhexlify(
            "0a500080000000000000000000000000000000000000000000000000000000000000000000d0"
        )
        msg = IPMessageRequest.build(
            dict(
                header=dict(
                    command=IPMessageCommand.toggle_keep_alive,
                    flags=dict(installer_mode=True),
                    cryptor_code="aes_256_ecb",
                ),
                payload=payload,
            ),
            password=self.connection.key,
        )
        await self.connection.send_raw_ip_message(msg)
        in_message = await self.connection.wait_for_ip_message()
        logger.debug(
            "Toggle keep alive response: {}".format(
                binascii.hexlify(in_message.payload)
            )
        )

    async def _send_upload_download_connection_command(self):
        # F3
        logger.debug("Sending upload download connection request")
        msg = IPMessageRequest.build(
            dict(
                header=dict(
                    command=IPMessageCommand.upload_download_connection,
                    flags=dict(installer_mode=True),
                    cryptor_code="aes_256_ecb",
                )
            ),
            password=self.connection.key,
        )
        await self.connection.send_raw_ip_message(msg)
        in_message = await self.connection.wait_for_ip_message()
        logger.debug(
            "Upload download connection response: {}".format(
                binascii.hexlify(in_message.payload)
            )
        )

    async def _send_keep_alive_command(self):
        # F2
        logger.debug("Sending keep alive request")
        msg = IPMessageRequest.build(
            dict(
                header=dict(
                    command=IPMessageCommand.keep_alive,
                    flags=dict(installer_mode=True),
                    cryptor_code="aes_256_ecb",
                ),
                payload=b"\x00\x00\x00\x00",
            ),
            password=self.connection.key,
        )
        await self.connection.send_raw_ip_message(msg)
        in_message = await self.connection.wait_for_ip_message()
        logger.debug(
            "Keep alive response: {}".format(binascii.hexlify(in_message.payload))
        )

    async def _authenticate_to_ip_module(self):
        logger.info("Authenticating with IP Module")

        self.connection.reset_key()
        msg = IPMessageRequest.build(
            dict(
                header=dict(
                    command=IPMessageCommand.connect,
                    # sub_command=3,
                    flags=dict(installer_mode=True),
                    cryptor_code="aes_256_ecb",
                ),
                payload=self.connection.key,
            ),
            password=self.connection.key,
        )
        await self.connection.send_raw_ip_message(msg)
        in_message = await self.connection.wait_for_ip_message()
        response = IPPayloadConnectResponse.parse(in_message.payload)

        if response.login_status != "success":
            logger.error(f"Error connecting to IP Module: {response.login_status}")

            if response.login_status == "invalid_password":
                raise PAICriticalException("Invalid IP Module password")
            else:
                raise ConnectToIpModuleFailed(f"Reason: {response.login_status}")

        logger.info(
            "Authentication Success. IP({}) Module version {:02x}, firmware: {}.{}, serial: {}".format(
                response.ip_type,
                response.hardware_version,
                response.ip_firmware_major,
                response.ip_firmware_minor,
                binascii.hexlify(response.ip_module_serial).decode("utf-8"),
            )
        )

        self.connection.set_key(response.key)
        self.ip_response_flags = in_message.header.flags



import binascii
import inspect
import logging
import typing
from abc import abstractmethod
from collections import defaultdict, namedtuple
from itertools import chain
from time import time

from construct import Construct, Container, EnumIntegerString

from paradox.config import config as cfg
from paradox.lib.utils import construct_free, sanitize_key

from ..lib import ps
from . import parsers

logger = logging.getLogger("PAI").getChild(__name__)

IndexAddress = namedtuple("IndexAddress", "idx address")


class Panel:
    mem_map = {}  # override in a subclass
    event_map = {}  # override in a subclass
    property_map = {}  # override in a subclass
    max_eeprom_response_data_length = 0  # override in a subclass
    status_request_addresses = []  # override in a subclass

    def __init__(self, core, variable_message_length=True):
        self.core = core
        self.settings = {}
        self.variable_message_length = variable_message_length

    def parse_message(self, message, direction="topanel") -> typing.Optional[Container]:
        if message is None or len(message) == 0:
            return None

        if direction == "topanel":
            if message[0] == 0x72 and message[1] == 0:
                return parsers.InitiateCommunication.parse(message)
            elif message[0] == 0x5F:
                return parsers.StartCommunication.parse(message)
        else:
            if message[0] == 0x72 and message[1] == 0xFF:
                return parsers.InitiateCommunicationResponse.parse(message)
            elif message[0] == 0x00 and message[4] > 0:
                return parsers.StartCommunicationResponse.parse(message)
            else:
                return None

    def get_message(self, name) -> Construct:
        clsmembers = dict(inspect.getmembers(parsers))
        if name in clsmembers:
            return clsmembers[name]
        else:
            raise ResourceWarning("{} parser not found".format(name))

    @staticmethod
    def get_error_message(error_code) -> str:
        # This is from EVO and may not apply to all panels

        error_str = str(error_code)
        if isinstance(error_code, EnumIntegerString):
            error_code = int(error_code)

        if error_code == 0x00:
            message = "Requested command did not work"
        elif error_code == 0x01:
            message = "User Code is invalid"
        elif error_code == 0x02:
            message = "Partition in code lockout (too many bad entries)"
        elif error_code == 0x05:
            message = "Panel will disconnect"
        elif error_code == 0x10:
            message = "Panel Not connected"
        elif error_code == 0x11:
            message = "Panel Already Connected"
        elif error_code == 0x12:
            message = "Invalid PC Password"
        elif error_code == 0x13:
            message = "Winload on phone line"
        elif error_code == 0x14:
            message = "Invalid Module address"
        elif error_code == 0x15:
            message = "Cannot write in RAM"
        elif error_code == 0x16:
            message = "Request to Upgrade Failed"
        elif error_code == 0x17:
            message = "Record number out of range"
        elif error_code == 0x19:
            message = "Invalid record type"
        elif error_code == 0x1A:
            message = "Multi-Bus not supported"
        elif error_code == 0x1B:
            message = "Incorrect number of users"
        elif error_code == 0x1C:
            message = "Invalid label number"
        else:
            message = error_str

        return message

    @staticmethod
    def encode_password(password) -> bytes:
        if password is None:
            password = 0

        if isinstance(password, int):
            password = str(password).zfill(4)
        elif isinstance(password, str):
            password = password.encode()

        return binascii.a2b_hex(password.zfill(4))

    async def load_memory(self):
        logger.info("Loading definitions")
        definitions = await self.load_definitions()
        ps.sendMessage("definitions_loaded", data=definitions)

        logger.info("Loading labels")
        labels = await self.load_labels()
        ps.sendMessage("labels_loaded", data=labels)

    async def load_definitions(self):
        if "definitions" not in self.mem_map:
            return {}
        logger.info("Updating Definitions from Panel")

        data = defaultdict(dict)
        try:
            def_parsers = self.get_message("DefinitionsParserMap")

            definitions = self.mem_map["definitions"]
            for elem_type in definitions:
                if elem_type not in def_parsers:
                    logger.warning("No parser for %s definitions", elem_type)
                    continue

                start_time = time()
                parser = def_parsers[elem_type]
                if isinstance(parser, typing.Callable):
                    parser = parser(self.settings)

                assert isinstance(parser, Construct)
                elem_def = definitions[elem_type]
                limits = cfg.LIMITS.get(elem_type)
                enabled_indexes = set()

                addresses = enumerate(
                    chain.from_iterable(elem_def["addresses"]), start=1
                )

                async for index, raw_data in self._eeprom_batch_reader(
                    addresses, parser.sizeof()
                ):
                    element = parser.parse(raw_data)
                    if cfg.LOGGING_DUMP_MESSAGES:
                        logger.debug(f"EEPROM parsed ({elem_type}/{index}): {element}")
                    if elem_def.get("bit_encoded"):
                        for elem_index, elem_data in element.items():
                            definition = elem_data.get("definition")
                            data_index = (index - 1) * len(element) + elem_index
                            data[elem_type][data_index] = elem_data
                            if definition != "disabled":
                                enabled_indexes.add(data_index)
                    else:
                        data[elem_type][index] = element
                        definition = element.get("definition")
                        if definition != "disabled":
                            enabled_indexes.add(index)

                if limits is None:
                    cfg.LIMITS[elem_type] = enabled_indexes
                else:
                    cfg.LIMITS[elem_type] = list(
                        set(cfg.LIMITS[elem_type]).intersection(enabled_indexes)
                    )

                logger.info(
                    f"{elem_type.title()} definitions loaded ({round(time() - start_time, 2)}s)"
                )

        except ResourceWarning:
            pass

        return construct_free(data)

    async def load_labels(self):
        logger.info("Updating Labels from Panel")

        data = defaultdict(dict)

        for elem_type in self.mem_map["labels"]:
            start_time = time()
            elem_def = self.mem_map["labels"][elem_type]

            addresses = enumerate(chain.from_iterable(elem_def["addresses"]), start=1)
            limits = cfg.LIMITS.get(elem_type)
            if limits is not None:
                addresses = iter((i, a) for i, a in addresses if i in limits)

            await self._load_labels(
                data, elem_type, addresses, label_offset=elem_def["label_offset"]
            )

            logger.info(
                f"{elem_type.title()} labels loaded ({round(time()-start_time, 2)}s): "
                + f"{', '.join([v['label'] for v in data[elem_type].values()])}"
            )

        return data

    async def _eeprom_read_address(self, address, length):
        args = dict(address=address, length=length)
        reply = await self.core.send_wait(
            self.get_message("ReadEEPROM"),
            args,
            reply_expected=lambda m: m.fields.value.po.command == 0x05
            and m.fields.value.address == address,
        )

        if reply is None:
            logger.error("Could not fully load labels")
            return

        return reply.fields.value.data

    async def _eeprom_batch_reader(self, addresses, field_length):
        if self.max_eeprom_response_data_length < 1:
            raise AssertionError("Wrong max_eeprom_response_data_length")

        batch = []
        while True:
            send_batch = None
            try:
                ia = IndexAddress(*next(addresses))
                batch_len = len(batch)
                if batch and (
                    (
                        batch[0].address + batch_len * field_length != ia.address
                    )  # Addresses are not sequential
                    or (
                        (batch_len + 1) * field_length
                        > self.max_eeprom_response_data_length
                    )  # one more field will not fit
                ):
                    send_batch = batch
                    batch = []

                batch.append(ia)
            except StopIteration:
                send_batch = batch
                batch = []

            if send_batch is not None:
                request_length = len(send_batch) * field_length
                if request_length == 0:
                    break
                else:
                    data = await self._eeprom_read_address(
                        send_batch[0].address, request_length
                    )
                    for i, ia2 in enumerate(send_batch, start=0):
                        yield ia2.idx, data[i * field_length : (i + 1) * field_length]

    async def _load_labels(
        self,
        data: dict,
        elem_type: str,
        addresses: typing.Iterator[typing.Tuple[int, int]],
        field_length=16,
        label_offset=0,
        template=None,
    ):
        """
        Load labels from panel

        :param data_dict: Dict to fill
        :param addresses: Addresses list with indexes
        :param field_length: Text field length
        :param label_offset: Label offset
        :param template: Default template
        :return:
        """
        element_dict = data[elem_type]

        if template is None:
            template = {}

        async for index, data in self._eeprom_batch_reader(addresses, field_length):
            b_label = data[label_offset : label_offset + field_length].strip(b"\0 ")

            label = b_label.replace(b"\0", b" ")

            try:
                label = label.decode(cfg.LABEL_ENCODING)
            except UnicodeDecodeError:
                logger.warning(
                    "Unable to properly decode label {} using the {} encoding.\n \
                    Specify a different encoding using the LABEL_ENCODING configuration option.".format(
                        b_label, cfg.LABEL_ENCODING
                    )
                )
                label = label.decode("utf-8", errors="ignore")

            properties = template.copy()
            properties["id"] = index
            properties["key"] = sanitize_key(label) or sanitize_key(
                f"{elem_type} {index}"
            )
            properties["label"] = label
            element_dict[index] = properties

    @abstractmethod
    def initialize_communication(self, password):
        raise NotImplementedError("override initialize_communication in a subclass")

    def get_status_requests(self) -> typing.Iterable[typing.Awaitable]:
        return (self.request_status(i) for i in self.status_request_addresses)

    @abstractmethod
    async def request_status(self, nr) -> typing.Optional[Container]:
        raise NotImplementedError("override request_status in a subclass")

    @staticmethod
    def handle_status(message: Container, parser_map):
        """Handle MessageStatus"""
        mvars = message.fields.value

        if mvars.address not in parser_map:
            logger.error(
                "Parser for memory address ({}) is not implemented. "
                "Skipping.".format(mvars.address)
            )
            return

        parser = parser_map[mvars.address]
        try:
            res = parser.parse(mvars.data)
            if cfg.LOGGING_DUMP_MESSAGES:
                logger.debug(f"Status parsed({mvars.address}): {res}")
            return res
        except:
            logger.exception(
                "Unable to parse RAM Status Block ({})".format(mvars.address)
            )
            return

    @abstractmethod
    def control_zones(self, zones, command) -> bool:
        raise NotImplementedError("override control_zones in a subclass")

    @abstractmethod
    def control_partitions(self, partitions, command) -> bool:
        raise NotImplementedError("override control_partitions in a subclass")

    @abstractmethod
    def control_outputs(self, outputs, command) -> bool:
        raise NotImplementedError("override control_outputs in a subclass")

    @abstractmethod
    def control_doors(self, doors, command) -> bool:
        raise NotImplementedError("override control_doors in a subclass")

    @abstractmethod
    def dump_memory(self, file, memory_type):
        raise NotImplementedError("override dump_memory in a subclass")

    @abstractmethod
    def send_panic(self, partition, panic_type, user_id):
        raise NotImplementedError("override send_panic in a subclass")



#!/usr/bin/env python3
import argparse
import binascii
from collections import OrderedDict

import yaml

from paradox.connections.ip.parsers import (IPMessageCommand, IPMessageRequest,
                                            IPMessageResponse, IPMessageType,
                                            IPPayloadConnectResponse)
from paradox.hardware import create_panel


def ordered_load(stream, Loader=yaml.loader.SafeLoader, object_pairs_hook=OrderedDict):
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
    )
    return yaml.load(stream, OrderedLoader)


class PayloadParser:
    def __init__(self):
        self.panel = create_panel(None)

    def parse(self, parsed):
        message_type = parsed.header.message_type
        if message_type == IPMessageType.serial_passthrough_request:
            self._parse_serial_passthrough_request(parsed)
        elif message_type == IPMessageType.serial_passthrough_response:
            self._parse_serial_passthrough_response(parsed)
        elif message_type == IPMessageType.ip_request:
            self._parse_ip_request(parsed)
        elif message_type == IPMessageType.ip_response:
            self._parse_ip_response(parsed)

    def _parse_ip_response(self, parsed):
        if parsed.header.command == IPMessageCommand.connect:
            print(IPPayloadConnectResponse.parse(parsed.payload))
        else:
            print(
                f"No parser for ip_response payload: {binascii.hexlify(parsed.payload)}"
            )

    def _parse_serial_passthrough_response(self, parsed):
        parsed_payload = self.panel.parse_message(parsed.payload, direction="frompanel")
        if parsed_payload is not None:
            if parsed_payload is not None:
                print(parsed_payload)
            else:
                print(
                    f"No parser for serial_passthrough_response payload: {binascii.hexlify(parsed.payload)}"
                )

            if parsed_payload.fields.value.po.command == 0:  # panel detection
                self.panel = create_panel(None, parsed_payload)
            if parsed_payload.fields.value.po.command == 5:  # eeprom/ram read
                self._parse_serial_passthrough_eeprom_read(parsed_payload)

    def _parse_serial_passthrough_eeprom_read(self, parsed_payload):
        if (
            "control" in parsed_payload.fields.value
            and parsed_payload.fields.value.control.ram_access
            and parsed_payload.fields.value.control._eeprom_address_bits == 0
            and parsed_payload.fields.value.bus_address == 0
        ):
            ram_address = parsed_payload.fields.value.address
            ram_parser = self.panel.get_message("RAMDataParserMap").get(ram_address)
            if ram_parser is not None:
                print(ram_parser.parse(parsed_payload.fields.value.data))
            else:
                print(
                    f"No parser for {ram_address} ram address, data: {binascii.hexlify(parsed_payload.fields.value.data)}"
                )

    def _parse_serial_passthrough_request(self, parsed):
        parsed_payload = self.panel.parse_message(parsed.payload)
        if parsed_payload is not None:
            print(parsed_payload)
        else:
            print(
                f"No parser for serial_passthrough_request payload: {binascii.hexlify(parsed.payload)}"
            )

    def _parse_ip_request(self, parsed):
        print(f"No parser for ip_request payload: {binascii.hexlify(parsed.payload)}")


def decrypt_file(file, password):
    try:
        data = ordered_load(file, yaml.loader.SafeLoader)
        parser = PayloadParser()

        n = 0
        for key, value in data.items():
            header = value[0:16]

            if "peer0_" in key:
                parsed = IPMessageRequest.parse(value, password=password)
            else:
                parsed = IPMessageResponse.parse(value, password=password)

            if (
                parsed.header.command == IPMessageCommand.connect
                and parsed.header.message_type == IPMessageType.ip_request
            ):
                assert password == parsed.payload, "Wrong decryption password"

            if (
                parsed.header.command == IPMessageCommand.connect
                and parsed.header.message_type == IPMessageType.ip_response
            ):
                password = parsed.payload[1:17]

            print(
                "PC->IP: " if "peer0_" in key else "IP->PC: ",
                f"header: {binascii.hexlify(header)}",
                f"body: {binascii.hexlify(parsed.payload)}",
                f"body_raw: {parsed.payload}",
            )

            print(parsed)
            parser.parse(parsed)

            if "peer1_" in key:
                print(
                    "----end %s-------------------------------------------------------------"
                    % key
                )
            n += 1

    except yaml.YAMLError as exc:
        print(exc)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file",
        type=argparse.FileType("r"),
        help="YAML file to parse. In wireshark right click on the first "
        "package, 'Follow->TCP Stream', 'Show and save data as': "
        "'YAML', copy contents to a file.",
    )
    parser.add_argument(
        "password",
        type=str,
        default="paradox",
        help="IP Module password for decryption",
    )

    args = parser.parse_args()

    decrypt_file(args.file, args.password.encode("utf8"))


if __name__ == "__main__":
    main()



from abc import abstractmethod

from construct import Container


class ConnectionHandler:
    @abstractmethod
    def on_message(self, raw: bytes):
        """
        Connection protocol will call this method when a serial pass through payload be received
        :param raw: bytes
        :return:
        """

    @abstractmethod
    def on_connection(self):
        """
        Connection protocol will call this method when a connection is established
        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def on_connection_loss(self):
        """
        Connection protocol will call this method when a connection was lost.
        :return:
        """
        raise NotImplementedError()


class IPConnectionHandler(ConnectionHandler):
    @abstractmethod
    def on_ip_message(self, container: Container):
        """
        Connection protocol will call this method upon receiving an ip150 message that is not a serial pass through.
        Will not be used with IP Bare and Serial connections.
        :param container: IPMessageResponse parse result
        :return:
        """



import codecs
import logging
import re
from importlib import import_module

logger = logging.getLogger("PAI").getChild(__name__)

codec_cache = {}


def paradox_codec_search(name: str):
    name = name.lower()
    match = re.match("^paradox-([a-z]{2})$", name)
    if match:
        enc = match.group(1)
        mod = codec_cache.get(enc)
        if mod is not None:
            return mod
        else:
            try:
                module = import_module("." + enc, "paradox.lib.encodings")
                mod = codec_cache[enc] = module.getregentry()
                logger.debug(f"Loaded {name} encoding")
                return mod
            except ImportError:
                pass


def register_encodings():
    codecs.register(paradox_codec_search)



from construct import (Adapter, Aligned, BitStruct, Bytes, Const, Default,
                       Enum, Flag, GreedyBytes, IfThenElse, Int8ub, Int16ub,
                       Int16ul, Padding, Pointer, Rebuild, Struct, len_, this)

from paradox.hardware.common import HexInt
from paradox.lib.crypto import decrypt, encrypt

IPMessageType = Enum(
    Int8ub,
    ip_response=0x1,
    serial_passthrough_response=0x2,
    ip_request=0x3,
    serial_passthrough_request=0x4,
)


IPMessageCommand = Enum(
    Int8ub,
    connect=0xF0,
    send_user_label=0xF1,
    keep_alive=0xF2,
    upload_download_connection=0xF3,
    upload_download_disconnection=0xF4,
    boot_loader=0xF5,
    web_page_connect=0xF6,
    web_page_disconnect=0xF7,
    toggle_keep_alive=0xF8,
    reset=0xF9,
    set_baud_rate=0xFA,
    multicommand=0xFB,
    single_panel=0xFC,
    unsupported_request=0xFD,
    boot_ip=0xFE,
    disconnect=0xFF,
    passthrough=0x00,
)


IPPayloadConnectResponse = Struct(
    "login_status"
    / Enum(
        Int8ub,
        success=0x00,
        invalid_password=0x01,
        user_already_connected=0x02,
        user_already_connected1=0x04,
    ),
    "key" / Bytes(16),
    "hardware_version" / Int16ub,
    "ip_firmware_major" / Default(HexInt, 5),
    "ip_firmware_minor" / Default(HexInt, 2),
    "ip_module_serial" / Bytes(4),
    "ip_type"
    / Default(
        Pointer(21, Enum(Int8ub, IP150=0x71, IP100=0x70)),
        lambda ctx: ctx.ip_module_serial[0],
    ),
)


class EncryptionAdapter(Adapter):
    def _decode(self, obj, context, path):
        try:
            return decrypt(obj, context._.password)[: context.header.length]
        except AttributeError:
            raise

    def _encode(self, obj, context, path):
        try:
            return encrypt(obj, context._.password)
        except AttributeError:
            raise


IPMessageRequest = Struct(
    "header"
    / Aligned(
        16,
        Struct(
            "sof" / Const(0xAA, Int8ub),
            "length"
            / Rebuild(
                Int16ul, lambda ctx: len(ctx._.payload) if "payload" in ctx._ else 0
            ),
            "message_type" / Default(IPMessageType, 0x03),
            "flags"
            / BitStruct(
                "bit8" / Default(Flag, False),
                "keep_alive" / Default(Flag, False),
                "live_events" / Default(Flag, False),
                "neware" / Default(Flag, False),
                "installer_mode" / Default(Flag, False),
                "bit3" / Default(Flag, False),
                "upload_download" / Default(Flag, False),
                "encrypt" / Default(Flag, True),
            ),
            "command" / Default(IPMessageCommand, 0x00),
            "sub_command" / Default(Int8ub, 0x00),
            "wt" / Default(Int8ub, 0x00),  # WT - 14
            "sb" / Default(Int8ub, 0x00),  # SB: 0
            "cryptor_code"
            / Default(Enum(Int8ub, none=0, aes_256_ecb=1, old_module=0xEE), "none"),
            "_not_used" / Padding(1, b"\xee"),
            "sequence_id" / Default(Int8ub, 0xEE),
        ),
        b"\xee",
    ),
    "payload"
    / Default(
        IfThenElse(
            this.header.flags.encrypt,
            EncryptionAdapter(Aligned(16, GreedyBytes, b"\xee")),
            Bytes(this.header.length),
        ),
        b"",
    ),
)


IPMessageResponse = Struct(
    "header"
    / Aligned(
        16,
        Struct(
            "sof" / Const(0xAA, Int8ub),
            "length"
            / Rebuild(
                Int16ul, lambda ctx: len(ctx._.payload) if "payload" in ctx._ else 0
            ),
            "message_type" / Default(IPMessageType, 0x01),
            "flags"
            / BitStruct(
                "bit8" / Default(Flag, False),
                "keep_alive" / Default(Flag, False),
                "live_events" / Default(Flag, False),
                "neware" / Default(Flag, False),
                "installer_mode" / Default(Flag, False),
                "bit3" / Default(Flag, False),
                "upload_download" / Default(Flag, False),
                "encrypt" / Default(Flag, True),
            ),
            "command" / Default(IPMessageCommand, 0x00),
            "sub_command" / Default(Int8ub, 0x00),
            "wt" / Default(Int8ub, 0x00),  # WT
            "sb" / Default(Int8ub, 0x03),  # SB
            "cryptor_code"
            / Default(Enum(Int8ub, none=0, aes_256_ecb=1, old_module=0xEE), "none"),
        ),
        b"\xee",
    ),
    "payload"
    / Default(
        IfThenElse(
            this.header.flags.encrypt,
            EncryptionAdapter(Aligned(16, GreedyBytes, b"\xee")),
            Bytes(this.header.length),
        ),
        b"",
    ),
)



# -*- coding: utf-8 -*-
import asyncio
import collections
import functools
import json
import re
import threading
import typing
from copy import deepcopy

from construct import Container, ListContainer

main_thread_loop = asyncio.get_event_loop()


def call_soon_in_main_loop(fn: typing.Union[typing.Callable, typing.Coroutine]) -> None:
    if threading.current_thread() is threading.main_thread():
        loop = asyncio.get_event_loop()
        if isinstance(fn, typing.Coroutine):
            loop.create_task(fn)
        else:
            loop.call_soon(fn)
    else:
        assert main_thread_loop.is_running()
        if isinstance(fn, typing.Coroutine):
            asyncio.run_coroutine_threadsafe(
                fn, loop=main_thread_loop
            )  # Returns concurrent.futures.Future
        else:
            main_thread_loop.call_soon_threadsafe(fn)


class JSONByteEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, bytes):
            return o.decode("utf-8")

        return super(JSONByteEncoder, self).default(o)


class SortableTuple(tuple):
    def __lt__(self, rhs):
        return self[0] < rhs[0]

    def __gt__(self, rhs):
        return self[0] > rhs[0]

    def __le__(self, rhs):
        return self[0] <= rhs[0]

    def __ge__(self, rhs):
        return self[0] >= rhs[0]


def deep_merge(*dicts, extend_lists=False, initializer=None):
    def merge_into(d1, d2):
        if d1 is None:
            return d2
        for key in d2:
            if key not in d1:  # key is missing
                d1[key] = deepcopy(d2[key])
            elif extend_lists and isinstance(d1[key], list):
                d1[key].extend(deepcopy(d2[key]))
            elif not isinstance(d1[key], dict):
                d1[key] = deepcopy(d2[key])
            else:
                d1[key] = merge_into(d1[key], d2[key])
        return d1

    return functools.reduce(merge_into, dicts, initializer)


re_sanitize_key = re.compile(r"\W")


def sanitize_key(key):
    if isinstance(key, int):
        return str(key)
    else:
        return re_sanitize_key.sub("_", key).strip("_")


def construct_free(container: Container):
    if isinstance(container, (Container, typing.Mapping)):
        return dict(
            (k, construct_free(v))
            for k, v in container.items()
            if not (isinstance(k, str) and k.startswith("_"))
        )
    elif isinstance(container, (ListContainer, typing.List)):
        return list(construct_free(v) for v in container)
    else:
        return container


class memoized(object):
    """From: https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
   Decorator. Caches a function's return value each time it is called.
   If called later with the same arguments, the cached value is returned
   (not reevaluated).
   """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if args in self.cache:
            return self.cache[args]
        else:
            if not isinstance(args, collections.Hashable):
                # uncacheable. a list, for instance.
                # better to not cache than blow up.
                return self.func(*args)

            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)



# -*- coding: utf-8 -*-

import asyncio
import datetime
import json
import logging
import os
from concurrent import futures

import serial_asyncio

from paradox.config import config as cfg
from paradox.connections.connection import ConnectionProtocol
from paradox.connections.handler import ConnectionHandler
from paradox.event import EventLevel, Notification
from paradox.interfaces.text.core import ConfiguredAbstractTextInterface
from paradox.lib import ps

# GSM interface.
# Only exposes critical status changes and accepts commands

logger = logging.getLogger("PAI").getChild(__name__)


class SerialConnectionProtocol(ConnectionProtocol):
    def __init__(self, handler: ConnectionHandler):
        super(SerialConnectionProtocol, self).__init__(handler)
        self.buffer = b""
        self.loop = asyncio.get_event_loop()
        self.last_message = b""

    def connection_made(self, transport):
        super(SerialConnectionProtocol, self).connection_made(transport)
        self.handler.on_connection()

    async def send_message(self, message):
        self.last_message = message
        self.transport.write(message + b"\r\n")

    def data_received(self, recv_data):
        self.buffer += recv_data
        logger.debug("BUFFER: {}".format(self.buffer))
        while len(self.buffer) >= 0:
            r = self.buffer.find(b"\r\n")
            # not found
            if r < 0:
                break

            # In the beginning
            if r == 0:
                self.buffer = self.buffer[2:]
                continue

            # Buffer is empty
            if len(self.buffer) == 0:
                return

            frame = self.buffer[:r]
            self.buffer = self.buffer[r:]
            # Ignore echoed bytes
            if self.last_message == frame:
                self.last_message = b""
            elif len(frame) > 0:
                self.loop.create_task(self.handler.on_message(frame))  # Callback

    def connection_lost(self, exc):
        logger.error("The serial port was closed")
        self.buffer = b""
        self.last_message = b""
        super(SerialConnectionProtocol, self).connection_lost(exc)


class SerialCommunication(ConnectionHandler):
    def __init__(self, loop, port, baud=9600, timeout=5, recv_callback=None):
        self.port_path = port
        self.baud = baud
        self.connected_future = None
        self.recv_callback = recv_callback
        self.loop = loop
        self.connected = False
        self.connection = None
        asyncio.set_event_loop(loop)
        self.queue = asyncio.Queue()

    def clear(self):
        self.queue = asyncio.Queue()

    def on_connection_loss(self):
        logger.error("Connection was lost")
        self.connected_future.set_result(False)
        self.connected = False

    def on_connection(self):
        logger.info("Serial port open")
        self.connected_future.set_result(True)
        self.connected = True

    def on_message(self, message: bytes):
        logger.debug("M->I: {}".format(message))

        if self.recv_callback is not None:
            return asyncio.get_event_loop().call_soon(
                self.recv_callback(message)
            )  # Callback
        else:
            return self.queue.put_nowait(message)

    def set_recv_callback(self, callback):
        self.recv_callback = callback

    def open_timeout(self):
        if self.connected_future.done():
            return

        logger.error("Serial Port Timeout")
        self.connected_future.set_result(False)
        self.connected = False

    def make_protocol(self):
        return SerialConnectionProtocol(self)

    async def write(self, message, timeout=15):
        logger.debug("I->M: {}".format(message))
        if self.connection is not None:
            await self.connection.send_message(message)
            return await asyncio.wait_for(self.queue.get(), timeout=5, loop=self.loop)

    async def read(self, timeout=5):
        if self.connection is not None:
            return await asyncio.wait_for(self.queue.get(), timeout=timeout)

    async def connect(self):
        logger.info("Connecting to serial port {}".format(self.port_path))

        self.connected_future = self.loop.create_future()
        self.loop.call_later(5, self.open_timeout)

        _, self.connection = await serial_asyncio.create_serial_connection(
            self.loop, self.make_protocol, self.port_path, self.baud
        )

        return await self.connected_future


class GSMTextInterface(ConfiguredAbstractTextInterface):
    """Interface Class using GSM"""

    def __init__(self, alarm):
        super().__init__(
            alarm,
            cfg.GSM_EVENT_FILTERS,
            cfg.GSM_ALLOW_EVENTS,
            cfg.GSM_IGNORE_EVENTS,
            cfg.GSM_MIN_EVENT_LEVEL,
        )

        self.port = None
        self.modem_connected = False
        self.loop = asyncio.new_event_loop()
        self.message_cmt = None

    def stop(self):
        """ Stops the GSM Interface Thread"""
        self.stop_running.set()

        self.loop.stop()
        super().stop()

        logger.debug("GSM Stopped")

    def write(self, message: str, expected: str = None) -> None:
        r = b""
        while r != expected:
            r = self.loop.run_until_complete(self.port.write(message))
            data = b""

            if r == b"ERROR":
                raise ("Got error from modem: {}".format(r))

            while r != expected:
                r = self.loop.run_until_complete(self.port.read())
                data += r + b"\n"

    def connect(self):
        logger.info(
            "Using {} at {} baud".format(cfg.GSM_MODEM_PORT, cfg.GSM_MODEM_BAUDRATE)
        )
        try:
            if not os.path.exists(cfg.GSM_MODEM_PORT):
                logger.error("Modem port ({}) not found".format(cfg.GSM_MODEM_PORT))
                return False

            self.port = SerialCommunication(
                self.loop, cfg.GSM_MODEM_PORT, cfg.GSM_MODEM_BAUDRATE, 5
            )

        except:
            logger.exception(
                "Could not open port {} for GSM modem".format(cfg.GSM_MODEM_PORT)
            )
            return False

        self.port.set_recv_callback(None)
        result = self.loop.run_until_complete(self.port.connect())

        if not result:
            logger.exception("Could not connect to GSM modem")
            return False

        try:
            self.write(b"AT", b"OK")  # Init
            self.write(b"ATE0", b"OK")  # Disable Echo
            self.write(b"AT+CMEE=2", b"OK")  # Increase verbosity
            self.write(b"AT+CMGF=1", b"OK")  # SMS Text mode
            self.write(b"AT+CFUN=1", b"OK")  # Enable modem
            self.write(
                b"AT+CNMI=1,2,0,0,0", b"OK"
            )  # SMS received only when modem enabled, Use +CMT with SMS, No Status Report,
            self.write(b"AT+CUSD=1", b"OK")  # Enable result code presentation

        except futures.TimeoutError as e:
            logger.error("No reply from modem")
            return False

        except:
            logger.exception("Modem connect error")
            return False

        self.port.set_recv_callback(
            self.data_received
        )  # Set recv callback to handle future messages

        logger.debug("Modem connected")
        self.modem_connected = True
        return True

    def _run(self):
        super(GSMTextInterface, self)._run()

        while not self.modem_connected and not self.stop_running.isSet():
            if not self.connect():
                logger.warning("Could not connect to modem")

            self.stop_running.wait(5)

        self.loop.run_forever()

        self.stop_running.wait()

    async def data_received(self, data: str) -> bool:
        logger.debug(f"Data Received: {data}")

        data = data.decode()

        if data.startswith("+CMT"):
            self.message_cmt = data
        elif self.message_cmt is not None:
            self.process_cmt(self.message_cmt, data)
            self.message_cmt = None
        elif data.startswith("+CUSD:"):
            self.process_cusd(data)

        return True

    def handle_message(self, timestamp: str, source: str, message: str) -> None:
        """ Handle GSM message. It should be a command """

        logger.debug("Received: {} {} {}".format(timestamp, source, message))

        if source in cfg.GSM_CONTACTS:
            future = asyncio.run_coroutine_threadsafe(
                self.handle_command(message), self.alarm.work_loop
            )
            ret = future.result(10)

            m = "GSM {}: {}".format(source, ret)
            logger.info(m)
        else:
            m = "GSM {} (UNK): {}".format(source, message)
            logger.warning(m)

        self.send_message(m, EventLevel.INFO)
        ps.sendNotification(
            Notification(sender=self.name, message=m, level=EventLevel.INFO)
        )

    def send_message(self, message: str, level: EventLevel) -> None:
        if self.port is None:
            logger.warning("GSM not available when sending message")
            return

        for dst in cfg.GSM_CONTACTS:
            data = b'AT+CMGS="%b"\x0d%b\x1a' % (dst.encode(), message.encode())

            try:
                future = asyncio.run_coroutine_threadsafe(
                    self.port.write(data), self.loop
                )
                result = future.result()
                logger.debug("SMS result: {}".format(result))
            except:
                logger.exception("ERROR sending SMS")

    def process_cmt(self, header: str, text: str) -> None:
        idx = header.find(" ")
        if idx <= 0:
            return

        tokens = json.loads(f"[{header[idx:]}]", strict=False)

        logger.debug("On {}, {} sent {}".format(tokens[2], tokens[0], text))
        self.handle_message(tokens[2], tokens[0], text)

    def process_cusd(self, message: str) -> None:
        idx = message.find(" ")
        if idx < 0:
            return

        tokens = json.loads(f"[{message[idx:]}]", strict=False)

        code = tokens[0]
        if code == 1:
            logger.info("Modem registered into network")
            ps.sendNotification(
                Notification(
                    sender=self.name,
                    message="Modem registered into network",
                    level=EventLevel.INFO,
                )
            )
        elif code == 4:
            logger.warning("CUSD code not supported")
            return
        else:
            return



import asyncio
import logging
from abc import abstractmethod
from typing import Any, Awaitable, Callable, List, Optional, Union

logger = logging.getLogger("PAI").getChild(__name__)


class Handler:
    def __init__(self, name=None):
        super().__init__()
        self.persistent = False
        self.name = name if name is not None else self.__class__.__name__

    def can_handle(self, data) -> bool:
        return True

    @abstractmethod
    async def __call__(self, data):
        """
        Handle message
        :param message:
        :return:
        """


class AlreadyHandledError(Exception):
    pass


class FutureHandler(Handler, asyncio.Future):
    def __init__(
        self, check_fn: Optional[Callable[[Any], bool]] = None, name=None,
    ):
        super().__init__()
        self.name = name if name is not None else self.__class__.__name__
        self._check_fn = check_fn

    def can_handle(self, data) -> bool:
        if self.done():
            raise AlreadyHandledError()
        if isinstance(self._check_fn, Callable):
            return self._check_fn(data)
        return True

    async def __call__(self, data):
        self.set_result(data)

        return data


class PersistentHandler(Handler):
    def __init__(
        self, callback: Callable[[Any], Union[None, Awaitable[None]]], name=None
    ):
        super(PersistentHandler, self).__init__(name)
        self._handle = callback
        self.persistent = True

    async def __call__(self, data):
        result = self._handle(data)

        if isinstance(result, Awaitable):
            return await result
        else:
            return result


class HandlerRegistry:
    def __init__(self, should_ignore_no_handlers=False):
        self._handlers: List[Handler] = []
        self._should_ignore_no_handlers = should_ignore_no_handlers

    def __len__(self):
        return len(self._handlers)

    def set_ignore_if_no_handlers(self, value):
        self._should_ignore_no_handlers = value

    def append(self, handler: Handler):
        assert isinstance(handler, Handler)
        self._handlers.append(handler)

    def remove(self, handler: Handler):
        assert isinstance(handler, Handler)
        self._handlers.remove(handler)

    def remove_by_name(self, name: str):
        to_remove = filter(lambda x: x.name == name, self._handlers)
        for handler in to_remove:
            self.remove(handler)

    async def wait_until_complete(self, handler: Handler, timeout=2):
        self.append(handler)
        try:
            return await asyncio.wait_for(handler, timeout=timeout)
        finally:
            assert not handler.persistent
            self.remove(handler)

    async def handle(self, data) -> None:
        """
        Find handler for inbound message and handle it.
        :param data:
        :return:
        """
        handled = False
        for handler in self._handlers:
            try:
                if handler.can_handle(data):
                    handled = True
                    await handler(data)
            except AlreadyHandledError:
                logger.error("Already handled")
            except:
                logger.exception("Exception caught during message handling")
                raise

        if not handled and not self._should_ignore_no_handlers:
            logger.error(
                "No handler for message {}\nDetail: {}".format(
                    data.fields.value.po.command, data
                )
            )



import logging

logger = logging.getLogger("PAI").getChild(__name__)


class PAIException(Exception):
    pass


class StatusRequestException(PAIException):
    pass


class PAIConnectionError(PAIException):
    pass


class ConnectToIpModuleFailed(PAIConnectionError):
    pass


class ConnectToSiteFailed(PAIConnectionError):
    pass


class StunSessionRefreshFailed(PAIConnectionError):
    pass


# CRITICAL Exceptions below that will stop PAI.
class PAICriticalException(PAIException):
    pass


class AuthenticationFailed(PAICriticalException):
    pass


class PanelNotDetected(PAICriticalException):
    pass


class SerialConnectionOpenFailed(PAICriticalException):
    pass


def async_loop_unhandled_exception_handler(loop, context):
    exception = context.get("exception")

    msg = context.get("exception", context["message"])

    task = context.get("task")
    if task:
        msg += ", task: %s" % task

    logger.exception(
        "Unhandled exception in async loop(%s): %s", loop, msg, exc_info=exception
    )
    loop.default_exception_handler(context)



from construct import (BitsInteger, BitStruct, Bytes, Const, Default, Enum,
                       Flag, Int8ub, Int16ub, Nibble, Padding, RawCopy, Struct)

from .common import (CommunicationSourceIDEnum, HexInt, PacketChecksum,
                     PacketLength, ProductIdEnum)

InitiateCommunication = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct("command" / Const(7, Nibble), "reserved0" / Const(2, Nibble)),
            "reserved1" / Padding(35),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

InitiateCommunicationResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / BitStruct("command" / Const(7, Nibble), "message_center" / Nibble),
            "new_protocol" / Const(0xFF, Int8ub),
            "protocol_id" / Int8ub,
            "protocol"
            / Struct("version" / Int8ub, "revision" / Int8ub, "build" / Int8ub),
            "family_id" / Int8ub,
            "product_id" / ProductIdEnum,
            "talker"
            / Enum(
                Int8ub, BOOT_LOADER=0, CONTROLLER_APPLICATION=1, MODULE_APPLICATION=2
            ),
            "application"
            / Struct("version" / HexInt, "revision" / HexInt, "build" / HexInt),
            "serial_number" / Bytes(4),
            "hardware" / Struct("version" / Int8ub, "revision" / Int8ub),
            "bootloader"
            / Struct(
                "version" / Int8ub,
                "revision" / Int8ub,
                "build" / Int8ub,
                "day" / Int8ub,
                "month" / Int8ub,
                "year" / Int8ub,
            ),
            "processor_id" / Int8ub,
            "encryption_id" / Int8ub,
            "reserved0" / Bytes(2),
            "label" / Bytes(8),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

StartCommunication = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x5F, Int8ub)),
            "validation" / Const(0x20, Int8ub),
            "_not_used0" / Padding(31),
            "source_id" / Default(CommunicationSourceIDEnum, 1),
            "user_id" / Struct("high" / Default(Int8ub, 0), "low" / Default(Int8ub, 0)),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

StartCommunicationResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "_not_used0" / Bytes(3),
            "product_id" / ProductIdEnum,
            "firmware"
            / Struct("version" / Int8ub, "revision" / Int8ub, "build" / Int8ub),
            "panel_id" / Int16ub,
            "_not_used1" / Bytes(5),
            "transceiver"
            / Struct(
                "firmware_build" / Int8ub,
                "family" / Int8ub,
                "firmware_version" / Int8ub,
                "firmware_revision" / Int8ub,
                "noise_floor_level" / Int8ub,
                "status"
                / BitStruct(
                    "_not_used" / BitsInteger(6),
                    "noise_floor_high" / Flag,
                    "constant_carrier" / Flag,
                ),
                "hardware_revision" / Int8ub,
            ),
            "_not_used2" / Bytes(14),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

CloseConnection = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x70, Int8ub)),
            "length" / PacketLength(Int8ub),
            "_not_used0" / Padding(34),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)



import logging

from paradox.interfaces import Interface

logger = logging.getLogger("PAI").getChild(__name__)


class InterfaceManager:
    def __init__(self, alarm, config=None):
        self.alarm = alarm
        self.conf = config
        self.interfaces = []

    def start(self):
        if self.conf.GSM_ENABLE:
            try:
                from paradox.interfaces.text.gsm import GSMTextInterface

                self.register(GSMTextInterface(self.alarm))
            except:
                logger.exception("Unable to start GSM Interface")

        # Load Signal service
        if self.conf.SIGNAL_ENABLE:
            try:
                from paradox.interfaces.text.signal import SignalTextInterface

                self.register(SignalTextInterface(self.alarm))
            except:
                logger.exception("Unable to start Signal Interface")

        # Load an interface for exposing data and accepting commands
        if self.conf.MQTT_ENABLE:
            try:
                from paradox.interfaces.mqtt.basic import BasicMQTTInterface

                self.register(BasicMQTTInterface(self.alarm))
            except:
                logger.exception("Unable to start MQTT Interface")

        if self.conf.MQTT_HOMEASSISTANT_AUTODISCOVERY_ENABLE:
            try:
                from paradox.interfaces.mqtt.homeassistant import (
                    HomeAssistantMQTTInterface,
                )

                self.register(HomeAssistantMQTTInterface(self.alarm))
            except:
                logger.exception("Unable to start HomeAssistant MQTT Interface")

        if self.conf.HOMEASSISTANT_NOTIFICATIONS_ENABLE:
            try:
                from paradox.interfaces.text.homeassistant_notifications import (
                    HomeAssistantNotificationsTextInterface,
                )

                self.register(HomeAssistantNotificationsTextInterface(self.alarm))
            except:
                logger.exception("Unable to start HomeAssistant MQTT Interface")

        # Load Pushbullet service
        if self.conf.PUSHBULLET_ENABLE:
            try:
                from paradox.interfaces.text.pushbullet import PushbulletTextInterface

                self.register(PushbulletTextInterface(self.alarm))
            except:
                logger.exception("Unable to start Pushbullet Interface")

        # Load Pushover service
        if self.conf.PUSHOVER_ENABLE:
            try:
                from paradox.interfaces.text.pushover import PushoverTextInterface

                self.register(PushoverTextInterface(self.alarm))
            except:
                logger.exception("Unable to start Pushover Interface")

        # Load IP Interface
        if self.conf.IP_INTERFACE_ENABLE:
            try:
                from paradox.interfaces.ip_interface.interface import IPInterface

                self.register(IPInterface(self.alarm))
            except:
                logger.exception("Unable to start IP Interface")

        # Load Dummy Interface
        if self.conf.DUMMY_INTERFACE_ENABLE:
            try:
                from paradox.interfaces.text.dummy import DummyInterface

                self.register(DummyInterface(self.alarm))
            except:
                logger.exception("Unable to start Dummy Interface")

    def register(self, interface: Interface):
        logger.debug("Registering {}".format(interface.name))
        interface.start()  # Starts interface thread

        self.interfaces.append(interface)

    def stop(self):
        logger.info("Stopping all interfaces")
        for interface in self.interfaces:
            try:
                logger.info(f"Stopping {interface.name}")
                interface.stop()
                interface.alarm = None
            except:
                logger.exception("Error stopping interface {}".format(interface.name))
        logger.debug("All Interfaces stopped")
        self.interfaces = []



# -*- coding: utf-8 -*-

import asyncio
import logging
from abc import ABC, abstractmethod

from construct import Container

from paradox.connections.connection import Connection
from paradox.connections.handler import IPConnectionHandler
from paradox.connections.ip.commands import IPModuleConnectCommand
from paradox.connections.ip.stun_session import StunSession
from paradox.connections.protocols import (IPConnectionProtocol,
                                           SerialConnectionProtocol)
from paradox.exceptions import PAICriticalException
from paradox.lib.handlers import FutureHandler, HandlerRegistry

logger = logging.getLogger("PAI").getChild(__name__)


class MultiAttemptConnection(Connection):
    async def connect(self) -> bool:
        tries = 1
        max_tries = 3
        self.connected = False

        while tries <= max_tries:
            logger.info("Connecting. Try %d/%d" % (tries, max_tries))
            try:
                await self._try_connect()
                return True
            except asyncio.TimeoutError as e:
                logger.error(
                    "Timeout while establishing connection (try %d/%d): %s"
                    % (tries, max_tries, str(e))
                )
            except OSError as e:
                logger.error(
                    "Connect failed (try %d/%d): %s" % (tries, max_tries, str(e))
                )
            except PAICriticalException:
                raise
            except:
                logger.exception(
                    "Unhandled exception while connecting (try %d/%d)"
                    % (tries, max_tries)
                )

            tries += 1

        return False

    @abstractmethod
    async def _try_connect(self):
        raise NotImplementedError("Implement in a subclass")


class BareIPConnection(MultiAttemptConnection):
    def __init__(self, host="127.0.0.1", port=10000):
        super().__init__()
        self.host = host
        self.port = port

    async def _try_connect(self):
        _, self._protocol = await self.loop.create_connection(
            self._make_protocol, host=self.host, port=self.port
        )

        self.connected = True

    def _make_protocol(self):
        return SerialConnectionProtocol(self)


class IPConnectionWithEncryption(MultiAttemptConnection, IPConnectionHandler, ABC):
    _protocol: IPConnectionProtocol

    def __init__(self, password=None):
        super().__init__()
        if isinstance(password, str):
            password = password.encode()
        self.password = password
        self.key = password

        self.ip_handler_registry = HandlerRegistry()

    def reset_key(self):
        self.set_key(self.password)

    def set_key(self, value):
        self.key = value
        self._protocol.key = value

    def on_ip_message(self, container: Container):
        return self.loop.create_task(self.ip_handler_registry.handle(container))

    async def wait_for_ip_message(self, timeout=2) -> Container:
        future = FutureHandler()
        return await self.ip_handler_registry.wait_until_complete(future, timeout)

    async def send_raw_ip_message(self, msg):
        self._protocol.send_raw(msg)

    def _make_protocol(self):
        return IPConnectionProtocol(self, self.key)


class LocalIPConnection(IPConnectionWithEncryption):
    def __init__(
        self, host, port, password,
    ):
        super().__init__(password)
        self.host = host
        self.port = port

    async def _try_connect(self) -> None:
        _, self._protocol = await self.loop.create_connection(
            self._make_protocol, host=self.host, port=self.port
        )

        await IPModuleConnectCommand(self).execute()

        self.connected = True


class StunIPConnection(IPConnectionWithEncryption):
    def __init__(self, site_id, email, panel_serial, password):
        super().__init__(password)

        self.stun_session = StunSession(site_id, email, panel_serial)

    def on_connection_loss(self):
        super().on_connection_loss()

        if self.stun_session is not None:
            self.stun_session.close()

    def write(self, data: bytes):
        """Write data to socket"""

        if self.stun_session is not None:
            self.stun_session.refresh_session_if_required()

        return super().write(data)

    async def _try_connect(self) -> None:
        await self.stun_session.connect()
        _, self._protocol = await self.loop.create_connection(
            self._make_protocol, sock=self.stun_session.get_socket()
        )

        await IPModuleConnectCommand(self).execute()

        self.connected = True



import codecs


class Codec(codecs.Codec):
    def encode(self, input, errors="strict"):
        return codecs.charmap_encode(input, errors, encoding_table)

    def decode(self, input, errors="strict"):
        return codecs.charmap_decode(input, errors, decoding_table)


class IncrementalEncoder(codecs.IncrementalEncoder):
    def encode(self, input, final=False):
        return codecs.charmap_encode(input, self.errors, encoding_table)[0]


class IncrementalDecoder(codecs.IncrementalDecoder):
    def decode(self, input, final=False):
        return codecs.charmap_decode(input, self.errors, decoding_table)[0]


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):
    pass


def getregentry():
    return codecs.CodecInfo(
        name="paradox-en",
        encode=Codec().encode,
        decode=Codec().decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
    )


# fmt: off
### Decoding Table
decoding_table = (
    "\x00"  # 0    0x00 -> NULL
    "\x01"  # 1    0x01 -> START OF HEADING
    "\x02"  # 2    0x02 -> START OF TEXT
    "\x03"  # 3    0x03 -> END OF TEXT
    "\x04"  # 4    0x04 -> END OF TRANSMISSION
    "\x05"  # 5    0x05 -> ENQUIRY
    "\x06"  # 6    0x06 -> ACKNOWLEDGE
    "\x07"  # 7    0x07 -> BELL
    "\x08"  # 8    0x08 -> BACKSPACE
    "\t"    # 9    0x09 -> HORIZONTAL TABULATION
    "\n"    # 10   0x0A -> LINE FEED
    "\x0b"  # 11   0x0B -> VERTICAL TABULATION
    "\x0c"  # 12   0x0C -> FORM FEED
    "\r"    # 13   0x0D -> CARRIAGE RETURN
    "\x0e"  # 14   0x0E -> SHIFT OUT
    "\x0f"  # 15   0x0F -> SHIFT IN
    "\x10"  # 16   0x10 -> DATA LINK ESCAPE
    "\x11"  # 17   0x11 -> DEVICE CONTROL ONE
    "\x12"  # 18   0x12 -> DEVICE CONTROL TWO
    "\x13"  # 19   0x13 -> DEVICE CONTROL THREE
    "\x14"  # 20   0x14 -> DEVICE CONTROL FOUR
    "\x15"  # 21   0x15 -> NEGATIVE ACKNOWLEDGE
    "\x16"  # 22   0x16 -> SYNCHRONOUS IDLE
    "\x17"  # 23   0x17 -> END OF TRANSMISSION BLOCK
    "\x18"  # 24   0x18 -> CANCEL
    "\x19"  # 25   0x19 -> END OF MEDIUM
    "\x1a"  # 26   0x1A -> SUBSTITUTE
    "\x1b"  # 27   0x1B -> ESCAPE
    "\x1c"  # 28   0x1C -> FILE SEPARATOR
    "\x1d"  # 29   0x1D -> GROUP SEPARATOR
    "\x1e"  # 30   0x1E -> RECORD SEPARATOR
    "\x1f"  # 31   0x1F -> UNIT SEPARATOR
    " "  # 32   0x20 -> SPACE
    "!"  # 33   0x21 -> EXCLAMATION MARK
    '"'  # 34   0x22 -> QUOTATION MARK
    "#"  # 35 0x23 -> NUMBER SIGN
    "$"  # 36   0x24 -> DOLLAR SIGN
    "%"  # 37   0x25 -> PERCENT SIGN
    "&"  # 38   0x26 -> AMPERSAND
    "'"  # 39   0x27 -> APOSTROPHE
    "("  # 40   0x28 -> LEFT PARENTHESIS
    ")"  # 41   0x29 -> RIGHT PARENTHESIS
    "*"  # 42   0x2A -> ASTERISK
    "+"  # 43   0x2B -> PLUS SIGN
    ","   # 44   0x2C -> COMMA
    "-"  # 45   0x2D -> HYPHEN-MINUS
    "."  # 46   0x2E -> FULL STOP
    "/"  # 47   0x2F -> SOLIDUS
    "0"  # 48   0x30 -> DIGIT ZERO
    "1"  # 49   0x31 -> DIGIT ONE
    "2"  # 50   0x32 -> DIGIT TWO
    "3"  # 51   0x33 -> DIGIT THREE
    "4"  # 52   0x34 -> DIGIT FOUR
    "5"  # 53   0x35 -> DIGIT FIVE
    "6"  # 54   0x36 -> DIGIT SIX
    "7"  # 55   0x37 -> DIGIT SEVEN
    "8"  # 56   0x38 -> DIGIT EIGHT
    "9"  # 57   0x39 -> DIGIT NINE
    ":"  # 58   0x3A -> COLON
    ";"  # 59   0x3B -> SEMICOLON
    "<"  # 60   0x3C -> LESS-THAN SIGN
    "="  # 61   0x3D -> EQUALS SIGN
    ">"  # 62   0x3E -> GREATER-THAN SIGN
    "?"  # 63   0x3F -> QUESTION MARK
    "@"  # 64   0x40 -> COMMERCIAL AT
    "A"  # 65   0x41 -> LATIN CAPITAL LETTER A
    "B"  # 66   0x42 -> LATIN CAPITAL LETTER B
    "C"  # 67   0x43 -> LATIN CAPITAL LETTER C
    "D"  # 68   0x44 -> LATIN CAPITAL LETTER D
    "E"  # 69   0x45 -> LATIN CAPITAL LETTER E
    "F"  # 70   0x46 -> LATIN CAPITAL LETTER F
    "G"  # 71   0x47 -> LATIN CAPITAL LETTER G
    "H"  # 72   0x48 -> LATIN CAPITAL LETTER H
    "I"  # 73   0x49 -> LATIN CAPITAL LETTER I
    "J"  # 74   0x4A -> LATIN CAPITAL LETTER J
    "K"  # 75   0x4B -> LATIN CAPITAL LETTER K
    "L"  # 76   0x4C -> LATIN CAPITAL LETTER L
    "M"  # 77   0x4D -> LATIN CAPITAL LETTER M
    "N"  # 78   0x4E -> LATIN CAPITAL LETTER N
    "O"  # 79   0x4F -> LATIN CAPITAL LETTER O
    "P"  # 80   0x50 -> LATIN CAPITAL LETTER P
    "Q"  # 81   0x51 -> LATIN CAPITAL LETTER Q
    "R"  # 82   0x52 -> LATIN CAPITAL LETTER R
    "S"  # 83   0x53 -> LATIN CAPITAL LETTER S
    "T"  # 84   0x54 -> LATIN CAPITAL LETTER T
    "U"  # 85   0x55 -> LATIN CAPITAL LETTER U
    "V"  # 86   0x56 -> LATIN CAPITAL LETTER V
    "W"  # 87   0x57 -> LATIN CAPITAL LETTER W
    "X"  # 88   0x58 -> LATIN CAPITAL LETTER X
    "Y"  # 89   0x59 -> LATIN CAPITAL LETTER Y
    "Z"  # 90   0x5A -> LATIN CAPITAL LETTER Z
    "["  # 91   0x5B -> LEFT SQUARE BRACKET
    "¥"  # 92   0x5C -> YEN
    "]"  # 93   0x5D -> RIGHT SQUARE BRACKET
    "^"  # 94   0x5E -> CIRCUMFLEX ACCENT
    "_"  # 95   0x5F -> LOW LINE
    "`"  # 96   0x60 -> GRAVE ACCENT
    "a"  # 97   0x61 -> LATIN SMALL LETTER A
    "b"  # 98   0x62 -> LATIN SMALL LETTER B
    "c"  # 99   0x63 -> LATIN SMALL LETTER C
    "d"  # 100  0x64 -> LATIN SMALL LETTER D
    "e"  # 101  0x65 -> LATIN SMALL LETTER E
    "f"  # 102  0x66 -> LATIN SMALL LETTER F
    "g"  # 103  0x67 -> LATIN SMALL LETTER G
    "h"  # 104  0x68 -> LATIN SMALL LETTER H
    "i"  # 105  0x69 -> LATIN SMALL LETTER I
    "j"  # 106  0x6A -> LATIN SMALL LETTER J
    "k"  # 107  0x6B -> LATIN SMALL LETTER K
    "l"  # 108  0x6C -> LATIN SMALL LETTER L
    "m"  # 109  0x6D -> LATIN SMALL LETTER M
    "n"  # 110  0x6E -> LATIN SMALL LETTER N
    "o"  # 111  0x6F -> LATIN SMALL LETTER O
    "p"  # 112  0x70 -> LATIN SMALL LETTER P
    "q"  # 113  0x71 -> LATIN SMALL LETTER Q
    "r"  # 114  0x72 -> LATIN SMALL LETTER R
    "s"  # 115  0x73 -> LATIN SMALL LETTER S
    "t"  # 116  0x74 -> LATIN SMALL LETTER T
    "u"  # 117  0x75 -> LATIN SMALL LETTER U
    "v"  # 118  0x76 -> LATIN SMALL LETTER V
    "w"  # 119  0x77 -> LATIN SMALL LETTER W
    "x"  # 120  0x78 -> LATIN SMALL LETTER X
    "y"  # 121  0x79 -> LATIN SMALL LETTER Y
    "z"  # 122  0x7A -> LATIN SMALL LETTER Z
    "{"  # 123  0x7B -> LEFT CURLY BRACKET
    "|"  # 124  0x7C -> VERTICAL LINE
    "}"  # 125  0x7D -> RIGHT CURLY BRACKET
    "→"  # 126  0x7E -> RIGHTWARDS ARROW
    "←"  # 127  0x7F ->  
    "Û"  # 128  0x80 -> 
    "Ù"  # 129  0x81 -> 
    "Ú"  # 130  0x82 -> 
    "Ü"  # 131  0x83 -> 
    "û"  # 132  0x84 -> 
    "ù"  # 133  0x85 -> 
    "ú"  # 134  0x86 -> 
    "Ô"  # 135  0x87 -> 
    "Ò"  # 136  0x88 -> 
    "Ó"  # 137  0x89 -> 
    "º"  # 138  0x8A -> 
    "ô"  # 139  0x8B -> 
    "ò"  # 140  0x8C -> 
    "ó"  # 141  0x8D -> 
    "ö"  # 142  0x8E -> 
    "¿"  # 143  0x8F -> 
    "Ê"  # 144  0x90 -> 
    "È"  # 145  0x91 -> 
    "É"  # 146  0x92 -> 
    "Ë"  # 147  0x93 -> 
    "ê"  # 148  0x94 -> 
    "è"  # 149  0x95 -> 
    "é"  # 150  0x96 -> 
    "ë"  # 151  0x97 -> 
    "Å"  # 152  0x98 -> 
    "Ä"  # 153  0x99 -> 
    "å"  # 154  0x9A -> 
    "â"  # 155  0x9B -> 
    "à"  # 156  0x9C -> 
    "á"  # 157  0x9D -> 
    "ä"  # 158  0x9E -> 
    "A"  # 159  0x9F -> 
    "a"  # 160  0xA0 -> 
    "Î"  # 161  0xA1 -> 
    "Ì"  # 162  0xA2 -> 
    "Í"  # 163  0xA3 -> 
    "Ï"  # 164  0xA4 -> 
    "¡"  # 165  0xA5 -> 
    "Ñ"  # 166  0xA6 -> 
    "ñ"  # 167  0xA7 -> 
    "N"  # 168  0xA8 ->  #Binary 168 what is that?
    "g"  # 169  0xA9 ->  #Binary 169 gram underlined?
    "g"  # 170  0xAA ->  #Binary 170 gram.
    "v"  # 171  0xAB ->  #Binary 171 no clue.
    "v"  # 172  0xAC -> 
    "w"  # 173  0xAD -> 
    " "  # 174  0xAE ->  #Binary 174 no idea what that is.
    "Æ"  # 175  0xAF -> 
    "§"  # 176  0xB0 -> 
    "±"  # 177  0xB1 -> 
    "i"  # 178  0xB2 -> 
    "↑"  # 179  0xB3 -> 
    "↓"  # 180  0xB4 -> 
    "↲"  # 181  0xB5 -> 
    "ƒ"  # 182  0xB6 -> 
    "£"  # 183  0xB7 -> 
    " "  # 184  0xB8 -> 
    "⤈"  # 185  0xB9 -> 
    "⤉"  # 186  0xBA -> 
    " "  # 187  0xBB -> 
    " "  # 188  0xBC -> 
    "½"  # 189  0xBD -> 
    "⅓"  # 190  0xBE -> 
    "¼"  # 191  0xBF -> 
    " "  # 192  0xC0 -> 
    " "  # 193  0xC1 -> 
    " "  # 194  0xC2 -> 
    " "  # 195  0xC3 -> 
    " "  # 196  0xC4 -> 
    " "  # 197  0xC5 -> 
    " "  # 198  0xC6 -> 
    " "  # 199  0xC7 -> 
    " "  # 200  0xC8 -> 
    " "  # 201  0xC9 -> 
    " "  # 202  0xCA -> 
    " "  # 203  0xCB -> 
    " "  # 204  0xCC -> 
    " "  # 205  0xCD -> 
    "Õ"  # 206  0xCE -> 
    "õ"  # 207  0xCF -> 
    " "  # 208  0xD0 -> 
    " "  # 209  0xD1 -> 
    " "  # 210  0xD2 -> 
    " "  # 211  0xD3 -> 
    " "  # 212  0xD4 -> 
    " "  # 213  0xD5 -> 
    " "  # 214  0xD6 -> 
    " "  # 215  0xD7 -> 
    " "  # 216  0xD8 -> 
    " "  # 217  0xD9 -> 
    "\\"  # 218  0xDA -> 
    "x"  # 219  0xDB -> 
    " "  # 220  0xDC -> 
    " "  # 221  0xDD -> 
    " "  # 222  0xDE -> 
    " "  # 223  0xDF ->
)

encoding_table = codecs.charmap_build(decoding_table)



import asyncio
import json
import logging
import typing
from collections import namedtuple

from paradox.config import config as cfg
from paradox.lib import ps
from paradox.lib.utils import sanitize_key

from ...data.model import DetectedPanel
from .core import AbstractMQTTInterface

logger = logging.getLogger("PAI").getChild(__name__)

PreparseResponse = namedtuple("preparse_response", "topics element content")


class HomeAssistantMQTTInterface(AbstractMQTTInterface):
    def __init__(self, alarm):
        super().__init__(alarm)
        self.armed = dict()
        self.partitions = {}
        self.zones = {}
        self.pgms = {}

        self.availability_topic = self.mqtt.availability_topic
        self.run_status_topic = self.mqtt.run_status_topic

        self.connected_future = (
            asyncio.Future()
        )  # TODO: do not create it, use some other
        panel_detected_future = asyncio.Future()
        first_status_update_future = asyncio.Future()

        def _ready_future_callback(x):
            self._publish_when_ready(
                panel_detected_future.result()["panel"],
                first_status_update_future.result()["status"],
            )

        ready_future = asyncio.ensure_future(
            asyncio.gather(
                self.connected_future, panel_detected_future, first_status_update_future
            )
        )
        ready_future.add_done_callback(_ready_future_callback)

        ps.subscribe(panel_detected_future, "panel_detected")
        ps.subscribe(self._handle_labels_loaded, "labels_loaded")
        ps.subscribe(first_status_update_future, "status_update")

    def on_connect(self, client, userdata, flags, result):
        # TODO: do not create connected_future, use some other
        if not self.connected_future.done():
            self.connected_future.set_result(True)

    def _handle_labels_loaded(self, data):
        partitions = data.get("partition", {})
        for k, v in partitions.items():
            p_data = {}
            p_data.update(v)
            self.partitions[k] = p_data

        self.zones = data.get("zone", {})
        self.pgms = data.get("pgm", {})

    def _publish_when_ready(self, panel: DetectedPanel, status):
        device = dict(
            manufacturer="Paradox",
            model=panel.model,
            identifiers=["Paradox", panel.model, panel.serial_number],
            name=panel.model,
            sw_version=panel.firmware_version,
        )

        self._publish_run_state_sensor(device, panel.serial_number)

        if "partition" in status:
            self._process_partition_statuses(
                status["partition"], device, panel.serial_number
            )
        if "zone" in status:
            self._process_zone_statuses(status["zone"], device, panel.serial_number)
        if "pgm" in status:
            self._process_pgm_statuses(status["pgm"], device, panel.serial_number)

    def _publish_run_state_sensor(self, device, device_sn):
        configuration_topic = "{}/sensor/{}/{}/config".format(
            cfg.MQTT_HOMEASSISTANT_DISCOVERY_PREFIX, device_sn, "run_status"
        )

        config = dict(
            name="Run status",
            unique_id="{}_partition_{}".format(device_sn, "run_status"),
            state_topic=self.run_status_topic,
            # availability_topic=self.availability_topic,
            device=device,
        )

        self.publish(configuration_topic, json.dumps(config), 0, cfg.MQTT_RETAIN)

    def _process_partition_statuses(self, partition_statuses, device, device_sn):
        for p_key, p_status in partition_statuses.items():
            if p_key not in self.partitions:
                continue
            partition = self.partitions[p_key]

            state_topic = "{}/{}/{}/{}/{}".format(
                cfg.MQTT_BASE_TOPIC,
                cfg.MQTT_STATES_TOPIC,
                cfg.MQTT_PARTITION_TOPIC,
                sanitize_key(partition["key"]),
                "current_state",
            )

            configuration_topic = "{}/alarm_control_panel/{}/{}/config".format(
                cfg.MQTT_HOMEASSISTANT_DISCOVERY_PREFIX,
                device_sn,
                sanitize_key(partition["key"]),
            )
            command_topic = "{}/{}/{}/{}".format(
                cfg.MQTT_BASE_TOPIC,
                cfg.MQTT_CONTROL_TOPIC,
                cfg.MQTT_PARTITION_TOPIC,
                sanitize_key(partition["key"]),
            )
            config = dict(
                name=partition["label"],
                unique_id="{}_partition_{}".format(device_sn, partition["key"]),
                command_topic=command_topic,
                state_topic=state_topic,
                availability_topic=self.availability_topic,
                device=device,
                payload_disarm="disarm",
                payload_arm_home="arm_stay",
                payload_arm_away="arm",
                payload_arm_night="arm_sleep",
            )

            self.publish(configuration_topic, json.dumps(config), 0, cfg.MQTT_RETAIN)

    def _process_zone_statuses(self, zone_statuses, device, device_sn):
        for z_key, p_status in zone_statuses.items():
            if z_key not in self.zones:
                continue

            zone = self.zones[z_key]

            open_topic = "{}/{}/{}/{}/{}".format(
                cfg.MQTT_BASE_TOPIC,
                cfg.MQTT_STATES_TOPIC,
                cfg.MQTT_ZONE_TOPIC,
                sanitize_key(zone["key"]),
                "open",
            )

            config = dict(
                name=zone["label"],
                unique_id="{}_zone_{}_open".format(device_sn, zone["key"]),
                state_topic=open_topic,
                device_class="motion",
                availability_topic=self.availability_topic,
                payload_on="True",
                payload_off="False",
                device=device,
            )

            configuration_topic = "{}/binary_sensor/{}/{}/config".format(
                cfg.MQTT_HOMEASSISTANT_DISCOVERY_PREFIX,
                device_sn,
                sanitize_key(zone["key"]),
            )

            self.publish(configuration_topic, json.dumps(config), 0, cfg.MQTT_RETAIN)

    def _process_pgm_statuses(self, pgm_statuses, device, device_sn):
        for pgm_key, p_status in pgm_statuses.items():
            if pgm_key not in self.pgms:
                continue

            pgm = self.pgms[pgm_key]

            on_topic = "{}/{}/{}/{}/{}".format(
                cfg.MQTT_BASE_TOPIC,
                cfg.MQTT_STATES_TOPIC,
                cfg.MQTT_OUTPUT_TOPIC,
                sanitize_key(pgm["key"]),
                "on",
            )

            command_topic = "{}/{}/{}/{}".format(
                cfg.MQTT_BASE_TOPIC,
                cfg.MQTT_CONTROL_TOPIC,
                cfg.MQTT_OUTPUT_TOPIC,
                sanitize_key(pgm["key"]),
            )

            config = dict(
                name=pgm["label"],
                unique_id="{}_pgm_{}_open".format(device_sn, pgm["key"]),
                state_topic=on_topic,
                command_topic=command_topic,
                availability_topic=self.availability_topic,
                state_on="True",
                state_off="False",
                payload_on="on",
                payload_off="off",
                device=device,
            )

            configuration_topic = "{}/switch/{}/{}/config".format(
                cfg.MQTT_HOMEASSISTANT_DISCOVERY_PREFIX,
                device_sn,
                sanitize_key(pgm["key"]),
            )

            self.publish(configuration_topic, json.dumps(config), 0, cfg.MQTT_RETAIN)



import codecs


class Codec(codecs.Codec):
    def encode(self, input, errors="strict"):
        return codecs.charmap_encode(input, errors, encoding_table)

    def decode(self, input, errors="strict"):
        return codecs.charmap_decode(input, errors, decoding_table)


class IncrementalEncoder(codecs.IncrementalEncoder):
    def encode(self, input, final=False):
        return codecs.charmap_encode(input, self.errors, encoding_table)[0]


class IncrementalDecoder(codecs.IncrementalDecoder):
    def decode(self, input, final=False):
        return codecs.charmap_decode(input, self.errors, decoding_table)[0]


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):
    pass


def getregentry():
    return codecs.CodecInfo(
        name="paradox-ru",
        encode=Codec().encode,
        decode=Codec().decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
    )


### Decoding Table

# fmt: off
### Decoding Table
decoding_table = (
    "\x00"  # 0    0x00 -> NULL
    "\x01"  # 1    0x01 -> START OF HEADING
    "\x02"  # 2    0x02 -> START OF TEXT
    "\x03"  # 3    0x03 -> END OF TEXT
    "\x04"  # 4    0x04 -> END OF TRANSMISSION
    "\x05"  # 5    0x05 -> ENQUIRY
    "\x06"  # 6    0x06 -> ACKNOWLEDGE
    "\x07"  # 7    0x07 -> BELL
    "\x08"  # 8    0x08 -> BACKSPACE
    "\t"    # 9    0x09 -> HORIZONTAL TABULATION
    "\n"    # 10   0x0A -> LINE FEED
    "\x0b"  # 11   0x0B -> VERTICAL TABULATION
    "\x0c"  # 12   0x0C -> FORM FEED
    "\r"    # 13   0x0D -> CARRIAGE RETURN
    "\x0e"  # 14   0x0E -> SHIFT OUT
    "\x0f"  # 15   0x0F -> SHIFT IN
    "\x10"  # 16   0x10 -> DATA LINK ESCAPE
    "\x11"  # 17   0x11 -> DEVICE CONTROL ONE
    "\x12"  # 18   0x12 -> DEVICE CONTROL TWO
    "\x13"  # 19   0x13 -> DEVICE CONTROL THREE
    "\x14"  # 20   0x14 -> DEVICE CONTROL FOUR
    "\x15"  # 21   0x15 -> NEGATIVE ACKNOWLEDGE
    "\x16"  # 22   0x16 -> SYNCHRONOUS IDLE
    "\x17"  # 23   0x17 -> END OF TRANSMISSION BLOCK
    "\x18"  # 24   0x18 -> CANCEL
    "\x19"  # 25   0x19 -> END OF MEDIUM
    "\x1a"  # 26   0x1A -> SUBSTITUTE
    "\x1b"  # 27   0x1B -> ESCAPE
    "\x1c"  # 28   0x1C -> FILE SEPARATOR
    "\x1d"  # 29   0x1D -> GROUP SEPARATOR
    "\x1e"  # 30   0x1E -> RECORD SEPARATOR
    "\x1f"  # 31   0x1F -> UNIT SEPARATOR
    " "     # 32   0x20 -> SPACE
    "!"     # 33   0x21 -> EXCLAMATION MARK
    '"'     # 34   0x22 -> QUOTATION MARK
    "#"  # 35   0x23 -> NUMBER SIGN
    "$"  # 36   0x24 -> DOLLAR SIGN
    "%"  # 37   0x25 -> PERCENT SIGN
    "&"  # 38   0x26 -> AMPERSAND
    "'"  # 39   0x27 -> APOSTROPHE
    "("  # 40   0x28 -> LEFT PARENTHESIS
    ")"  # 41   0x29 -> RIGHT PARENTHESIS
    "*"  # 42   0x2A -> ASTERISK
    "+"  # 43   0x2B -> PLUS SIGN
    ","  # 44   0x2C -> COMMA
    "-"  # 45   0x2D -> HYPHEN-MINUS
    "."  # 46   0x2E -> FULL STOP
    "/"  # 47   0x2F -> SOLIDUS
    "0"  # 48   0x30 -> DIGIT ZERO
    "1"  # 49   0x31 -> DIGIT ONE
    "2"  # 50   0x32 -> DIGIT TWO
    "3"  # 51   0x33 -> DIGIT THREE
    "4"  # 52   0x34 -> DIGIT FOUR
    "5"  # 53   0x35 -> DIGIT FIVE
    "6"  # 54   0x36 -> DIGIT SIX
    "7"  # 55   0x37 -> DIGIT SEVEN
    "8"  # 56   0x38 -> DIGIT EIGHT
    "9"  # 57   0x39 -> DIGIT NINE
    ":"  # 58   0x3A -> COLON
    ";"  # 59   0x3B -> SEMICOLON
    "<"  # 60   0x3C -> LESS-THAN SIGN
    "="  # 61   0x3D -> EQUALS SIGN
    ">"  # 62   0x3E -> GREATER-THAN SIGN
    "?"  # 63   0x3F -> QUESTION MARK
    "@"  # 64   0x40 -> COMMERCIAL AT
    "A"  # 65   0x41 -> LATIN CAPITAL LETTER A
    "B"  # 66   0x42 -> LATIN CAPITAL LETTER B
    "C"  # 67   0x43 -> LATIN CAPITAL LETTER C
    "D"  # 68   0x44 -> LATIN CAPITAL LETTER D
    "E"  # 69   0x45 -> LATIN CAPITAL LETTER E
    "F"  # 70   0x46 -> LATIN CAPITAL LETTER F
    "G"  # 71   0x47 -> LATIN CAPITAL LETTER G
    "H"  # 72   0x48 -> LATIN CAPITAL LETTER H
    "I"  # 73   0x49 -> LATIN CAPITAL LETTER I
    "J"  # 74   0x4A -> LATIN CAPITAL LETTER J
    "K"  # 75   0x4B -> LATIN CAPITAL LETTER K
    "L"  # 76   0x4C -> LATIN CAPITAL LETTER L
    "M"  # 77   0x4D -> LATIN CAPITAL LETTER M
    "N"  # 78   0x4E -> LATIN CAPITAL LETTER N
    "O"  # 79   0x4F -> LATIN CAPITAL LETTER O
    "P"  # 80   0x50 -> LATIN CAPITAL LETTER P
    "Q"  # 81   0x51 -> LATIN CAPITAL LETTER Q
    "R"  # 82   0x52 -> LATIN CAPITAL LETTER R
    "S"  # 83   0x53 -> LATIN CAPITAL LETTER S
    "T"  # 84   0x54 -> LATIN CAPITAL LETTER T
    "U"  # 85   0x55 -> LATIN CAPITAL LETTER U
    "V"  # 86   0x56 -> LATIN CAPITAL LETTER V
    "W"  # 87   0x57 -> LATIN CAPITAL LETTER W
    "X"  # 88   0x58 -> LATIN CAPITAL LETTER X
    "Y"  # 89   0x59 -> LATIN CAPITAL LETTER Y
    "Z"  # 90   0x5A -> LATIN CAPITAL LETTER Z
    "["  # 91   0x5B -> LEFT SQUARE BRACKET
    "¥"  # 92   0x5C -> YEN
    "]"  # 93   0x5D -> RIGHT SQUARE BRACKET
    "^"  # 94   0x5E -> CIRCUMFLEX ACCENT
    "_"  # 95   0x5F -> LOW LINE
    "`"  # 96   0x60 -> GRAVE ACCENT
    "a"  # 97   0x61 -> LATIN SMALL LETTER A
    "b"  # 98   0x62 -> LATIN SMALL LETTER B
    "c"  # 99   0x63 -> LATIN SMALL LETTER C
    "d"  # 100  0x64 -> LATIN SMALL LETTER D
    "e"  # 101  0x65 -> LATIN SMALL LETTER E
    "f"  # 102  0x66 -> LATIN SMALL LETTER F
    "g"  # 103  0x67 -> LATIN SMALL LETTER G
    "h"  # 104  0x68 -> LATIN SMALL LETTER H
    "i"  # 105  0x69 -> LATIN SMALL LETTER I
    "j"  # 106  0x6A -> LATIN SMALL LETTER J
    "k"  # 107  0x6B -> LATIN SMALL LETTER K
    "l"  # 108  0x6C -> LATIN SMALL LETTER L
    "m"  # 109  0x6D -> LATIN SMALL LETTER M
    "n"  # 110  0x6E -> LATIN SMALL LETTER N
    "o"  # 111  0x6F -> LATIN SMALL LETTER O
    "p"  # 112  0x70 -> LATIN SMALL LETTER P
    "q"  # 113  0x71 -> LATIN SMALL LETTER Q
    "r"  # 114  0x72 -> LATIN SMALL LETTER R
    "s"  # 115  0x73 -> LATIN SMALL LETTER S
    "t"  # 116  0x74 -> LATIN SMALL LETTER T
    "u"  # 117  0x75 -> LATIN SMALL LETTER U
    "v"  # 118  0x76 -> LATIN SMALL LETTER V
    "w"  # 119  0x77 -> LATIN SMALL LETTER W
    "x"  # 120  0x78 -> LATIN SMALL LETTER X
    "y"  # 121  0x79 -> LATIN SMALL LETTER Y
    "z"  # 122  0x7A -> LATIN SMALL LETTER Z
    " "  # 123  0x7B -> LEFT CURLY BRACKET
    " "  # 124  0x7C -> VERTICAL LINE
    " "  # 125  0x7D -> RIGHT CURLY BRACKET
    "↲"  # 126  0x7E -> 
    " "  # 127  0x7F ->
    " "  # 128  0x80 ->
    " "  # 129  0x81 ->
    " "  # 130  0x82 ->
    " "  # 131  0x83 ->
    " "  # 132  0x84 ->
    " "  # 133  0x85 ->
    " "  # 134  0x86 ->
    " "  # 135  0x87 ->
    " "  # 136  0x88 ->
    " "  # 137  0x89 ->
    " "  # 138  0x8A ->
    " "  # 139  0x8B ->
    " "  # 140  0x8C ->
    " "  # 141  0x8D ->
    " "  # 142  0x8E ->
    " "  # 143  0x8F ->
    " "  # 144  0x90 ->
    " "  # 145  0x91 ->
    " "  # 146  0x92 ->
    " "  # 147  0x93 ->
    " "  # 148  0x94 ->
    " "  # 149  0x95 ->
    " "  # 150  0x96 ->
    " "  # 151  0x97 ->
    " "  # 152  0x98 ->
    " "  # 153  0x99 ->
    " "  # 154  0x9A ->
    " "  # 155  0x9B ->
    " "  # 156  0x9C ->
    " "  # 157  0x9D ->
    " "  # 158  0x9E ->
    " "  # 159  0x9F ->
    "Б"  # 160  0xA0 ->
    "Г"  # 161  0xA1 ->
    "Ё"  # 162  0xA2 ->
    "Ж"  # 163  0xA3 ->
    "З"  # 164  0xA4 ->
    "И"  # 165  0xA5 ->
    "Й"  # 166  0xA6 ->
    "Л"  # 167  0xA7 ->
    "П"  # 168  0xA8 ->
    "У"  # 169  0xA9 ->
    "Ф"  # 170  0xAA ->
    "Ч"  # 171  0xAB ->
    "Ш"  # 172  0xAC ->
    "Ъ"  # 173  0xAD ->
    "Ы"  # 174  0xAE ->
    "Э"  # 175  0xAF ->
    "Ю"  # 176  0xB0 ->
    "Я"  # 177  0xB1 ->
    "б"  # 178  0xB2 ->
    "в"  # 179  0xB3 ->
    "г"  # 180  0xB4 ->
    "ё"  # 181  0xB5 ->
    "ж"  # 182  0xB6 ->
    "з"  # 183  0xB7 ->
    "и"  # 184  0xB8 ->
    "й"  # 185  0xB9 ->
    "к"  # 186  0xBA ->
    "л"  # 187  0xBB ->
    "м"  # 188  0xBC ->
    "н"  # 189  0xBD ->
    "п"  # 190  0xBE ->
    "т"  # 191  0xBF ->
    "ч"  # 192  0xC0 ->
    "ш"  # 193  0xC1 ->
    "ъ"  # 194  0xC2 ->
    "ы"  # 195  0xC3 ->
    "ь"  # 196  0xC4 ->
    "э"  # 197  0xC5 ->
    "ю"  # 198  0xC6 ->
    "я"  # 199  0xC7 ->
    " "  # 200  0xC8 ->
    " "  # 201  0xC9 ->
    " "  # 202  0xCA ->
    " "  # 203  0xCB ->
    " "  # 204  0xCC ->
    " "  # 205  0xCD ->
    " "  # 206  0xCE ->
    " "  # 207  0xCF ->
    " "  # 208  0xD0 ->
    " "  # 209  0xD1 ->
    " "  # 210  0xD2 ->
    " "  # 211  0xD3 ->
    " "  # 212  0xD4 ->
    " "  # 213  0xD5 ->
    " "  # 214  0xD6 ->
    " "  # 215  0xD7 ->
    " "  # 216  0xD8 ->
    " "  # 217  0xD9 ->
    " "  # 218  0xDA ->
    " "  # 219  0xDB ->
    " "  # 220  0xDC ->
    " "  # 221  0xDD ->
    " "  # 222  0xDE ->
    " "  # 223  0xDF ->
    "Д"  # 224  0xE0 ->
    "Ц"  # 225  0xE1 ->
    "Щ"  # 226  0xE2 ->
    "д"  # 227  0xE3 ->
    "ф"  # 228  0xE4 ->
    "ц"  # 229  0xE5 ->
    "щ"  # 230  0xE6 ->
    " "  # 231  0xE7 ->
    " "  # 232  0xE8 ->
    " "  # 233  0xE9 ->
    " "  # 234  0xEA ->
    " "  # 235  0xEB ->
    " "  # 236  0xEC ->
    " "  # 237  0xED ->
    " "  # 238  0xEE ->
    " "  # 239  0xEF ->
    " "  # 240  0xF0 ->
    " "  # 241  0xF1 ->
    " "  # 242  0xF2 ->
    " "  # 243  0xF3 ->
    " "  # 244  0xF4 ->
    " "  # 245  0xF5 ->
    " "  # 246  0xF6 ->
    " "  # 247  0xF7 ->
    " "  # 248  0xF8 ->
    " "  # 249  0xF9 ->
    " "  # 250  0xFA ->
    " "  # 251  0xFB ->
    " "  # 252  0xFC ->
    " "  # 253  0xFD ->
    " "  # 254  0xFE ->
    " "  # 255  0xFF ->
)

encoding_table = codecs.charmap_build(decoding_table)



from construct import Checksum, Enum, ExprAdapter, Int8ub, Rebuild, this


def calculate_checksum(message):
    r = 0
    for c in message:
        r += c
    r = r % 256
    return bytes([r])


def _hex_to_int(obj, path):
    try:
        return int(hex(obj)[2:], 10)
    except ValueError:
        return -1


def PacketLength(subcons):
    return Rebuild(
        subcons,
        lambda x: x._root._subcons.fields.sizeof() + x._root._subcons.checksum.sizeof(),
    )


def PacketChecksum(subcons):
    return Checksum(subcons, lambda data: calculate_checksum(data), this.fields.data)


HexInt = ExprAdapter(Int8ub, _hex_to_int, lambda obj, path: int(str(obj), 16))

ProductIdEnum = Enum(
    Int8ub,
    DIGIPLEX_v13=0,
    DIGIPLEX_v2=1,
    DIGIPLEX_NE=2,
    DIGIPLEX_EVO_48=3,
    DIGIPLEX_EVO_96=4,
    DIGIPLEX_EVO_192=5,
    DIGIPLEX_EVO_HD=7,
    SPECTRA_SP5500=21,
    SPECTRA_SP6000=22,
    SPECTRA_SP7000=23,
    SPECTRA_SP4000=26,
    SPECTRA_SP65=27,
    MAGELLAN_MG5000=64,
    MAGELLAN_MG5050=65,
)

CommunicationSourceIDEnum = Enum(
    Int8ub,
    NonValid_Source=0,
    Winload_Direct=1,
    Winload_IP=2,
    Winload_GSM=3,
    Winload_Dialer=4,
    NeWare_Direct=5,
    NeWare_IP=6,
    NeWare_GSM=7,
    NeWare_Dialer=8,
    IP_Direct=9,
    VDMP3_Direct=10,
    VDMP3_GSM=11,
)
